"""tipsy
=====

Implements classes and functions for handling tipsy files.  You rarely
need to access this module directly as it will be invoked
automatically via pynbody.load.

**Input**:

*filename*: file name string

**Optional Keywords**:

*paramfile*: string specifying the parameter file to load. If not
specified, the loader will look for a file `*.param` in the current and
parent directories.

"""

import copy
import glob
import logging
import math
import os
import struct
import sys
import warnings

import numpy as np

from .. import array, chunk, config, config_parser, family, units, util
from . import SimSnap, nchilada

logger = logging.getLogger('pynbody.snapshot.tipsy')


class TipsySnap(SimSnap):
    _basic_loadable_keys = {family.dm: {'phi', 'pos', 'eps', 'mass', 'vel'},
                            family.gas: {'phi', 'temp', 'pos', 'metals', 'eps',
                                             'mass', 'rho', 'vel'},
                            family.star: {'phi', 'tform', 'pos', 'metals',
                                              'eps', 'mass', 'vel'},
                            None: {'phi', 'pos', 'eps', 'mass', 'vel'}}

    def __init__(self, filename, **kwargs):

        global config

        super().__init__()

        only_header = kwargs.get('only_header', False)
        if only_header:
            warnings.warn(
                "only_header kwarg is deprecated: all loading in TipsySnap is now lazy by default", RuntimeWarning)

        must_have_paramfile = kwargs.get('must_have_paramfile', False)
        take = kwargs.get('take', None)
        verbose = kwargs.get('verbose', config['verbose'])

        self.partial_load = take is not None

        self._filename = str(util.cutgz(filename))

        if not only_header:
            logger.info("Loading %s", filename)
        with util.open_(filename, 'rb') as f:
            t, n, ndim, ng, nd, ns = struct.unpack("diiiii", f.read(28))
            if (ndim > 3 or ndim < 1):
                self._byteswap = True
                f.seek(0)
                t, n, ndim, ng, nd, ns = struct.unpack(">diiiii", f.read(28))
            else:
                self._byteswap = False

            assert ndim == 3

            self._header_t = t

            f.read(4)

        disk_family_slice = dict({family.gas: slice(0, ng),
                                  family.dm: slice(ng, nd + ng),
                                  family.star: slice(nd + ng, ng + nd + ns)})

        self._load_control = chunk.LoadControl(disk_family_slice, 10240, take)

        self._family_slice = self._load_control.mem_family_slice
        self._num_particles = self._load_control.mem_num_particles

        self._paramfilename = kwargs.get('paramfile', None)

        self._decorate()

        # describe the file structure as list of (num_parts, [list_of_properties])
        # by default all fields are floats -- we look at the param file to determine
        # whether we should expect some doubles

        if self._paramfile.get('bDoublePos', 0):
            ptype = 'd'
        else:
            ptype = 'f'

        if self._paramfile.get('bDoubleVel', 0):
            vtype = 'd'
        else:
            vtype = 'f'

        self._g_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "rho", "temp", "eps", "metals", "phi"),
                                  'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f', 'f', 'f', 'f')})
        self._d_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "eps", "phi"),
                                  'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f')})
        self._s_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "metals", "tform", "eps", "phi"),
                                  'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f', 'f', 'f')})

        if 'dKpcUnit' not in self._paramfile:
            if must_have_paramfile:
                raise RuntimeError("Could not find .param file for this run. Place it in the run's directory or parent directory.")
            else:
                warnings.warn(
                    "No readable param file in the run directory or parent directory: using defaults.", RuntimeWarning)
                self._file_units_system = [
                    units.Unit(x) for x in ('G', '1 kpc', '1e10 Msol')]

        time_unit = None
        try:
            time_unit = self.infer_original_units('yr')
        except units.UnitsException:
            pass

        if time_unit is not None:
            self.properties['time'] *= time_unit

    def _load_main_file(self):

        logger.info("Loading data from main file %s", self._filename)

        f = util.open_(self._filename, 'rb')
        f.seek(32)

        write = []

        for w, ndim in ("pos", 3), ("vel", 3), ("mass", 1), ("eps", 1), ("phi", 1):
            if w not in list(self.keys()):
                self._create_array(w, ndim, zeros=False)
                write.append(w)

        for w in "rho", "temp":
            if w not in list(self.gas.keys()):
                self.gas._create_array(w, zeros=False)
                write.append(w)

        if ("metals" not in list(self.gas.keys())) and ("metals" not in list(self.star.keys())):
            self.gas._create_array("metals", zeros=False)
            self.star._create_array("metals", zeros=False)
            write.append("metals")

        if "tform" not in list(self.star.keys()):
            self.star._create_array("tform", zeros=False)
            write.append("tform")

        if "temp" in write:
            self.gas["temp"].units = "K"

        for k in "pos", "vel", "mass", "eps", "phi":
            if k in write:
                self[k].set_default_units(quiet=True)

        # only do this for cosmo runs
        if "phi" in write and 'h' in self.properties:
            self['phi'].units = self['phi'].units * units.a ** -3  # messy :-(

        for k in "rho", "temp", "metals":
            if k in write:
                self.gas[k].set_default_units(quiet=True)

        for k in "metals", "tform":
            if k in write:
                self.star[k].set_default_units(quiet=True)

        if "pos" in write:
            write += ['x', 'y', 'z']

        if "vel" in write:
            write += ['vx', 'vy', 'vz']

        max_item_size = max(
            q.itemsize for q in (self._g_dtype, self._d_dtype, self._s_dtype))
        tbuf = bytearray(max_item_size * 10240)

        for fam, dtype in ((family.gas, self._g_dtype), (family.dm, self._d_dtype), (family.star, self._s_dtype)):
            self_fam = self[fam]
            st_len = dtype.itemsize
            for readlen, buf_index, mem_index in self._load_control.iterate([fam], [fam], multiskip=True):
                # Read in the block

                if mem_index is None:
                    f.seek(st_len * readlen, 1)
                    continue

                buf = np.frombuffer(f.read(st_len * readlen), dtype=dtype)

                if self._byteswap:
                    buf = buf.byteswap()

                if mem_index is not None:
                    # Copy into the correct arrays
                    for name in dtype.names:
                        if name in write:
                            self_fam[name][mem_index] = buf[name][buf_index]

        f.close()

    def _update_loadable_keys(self):
        def is_readable_array(x):
            try:
                with util.open_(x, 'r') as f:
                    return int(f.readline()) == len(self)
            except ValueError:
                # could be a binary file
                with util.open_(x, 'rb') as f:
                    header = f.read(4)
                if len(header) != 4:
                    return False

                if self._byteswap:
                    buflen = struct.unpack(">i", header)[0]
                else:
                    buflen = struct.unpack("i", header)[0]

                ourlen_1 = (self._load_control.disk_num_particles)& 0xffffffff
                ourlen_3 = (self._load_control.disk_num_particles*3)& 0xffffffff

                if buflen == ourlen_1:  # it's a vector
                    return True
                elif buflen == ourlen_3:  # it's an array
                    return True
                else:
                    return False

            except OSError:
                return False

        fs = list(map(util.cutgz, glob.glob(self._filename + ".*")))
        res = [q[len(self._filename) + 1:] for q in list(filter(is_readable_array, fs))]

        # Create an empty dictionary of sets to store the loadable
        # arrays for each family
        rdict = {x: set() for x in self.families()}
        rdict.update({a: copy.copy(b)
                          for a, b in self._basic_loadable_keys.items() if a is not None})
        # Now work out which families can load which arrays
        # according to the stored metadata
        for r in res:
            fams = self._get_loadable_array_metadata(r)[1]
            for x in fams or list(rdict.keys()):
                rdict[x].add(r)

        self._loadable_keys_registry = rdict

    def loadable_keys(self, fam=None):
        """Produce and return a list of loadable arrays for this TIPSY file."""
        if len(self._loadable_keys_registry) == 0:
            self._update_loadable_keys()

        if fam is not None:
            # Return what is loadable for this family
            return list(self._loadable_keys_registry[fam])
        else:
            # Return what is loadable to all families
            return list(set.intersection(*list(self._loadable_keys_registry.values())))

    def _update_snapshot(self, arrays, filename=None, fam_out=[family.gas, family.dm, family.star]):
        """
        Write a TIPSY file, but only updating the requested information and leaving the
        rest intact on disk.
        """

        if self.partial_load:
            raise RuntimeError("Writing back to partially loaded files not yet supported")

        global config

        # make arrays be a list
        if isinstance(arrays, str):
            arrays = [arrays]

        # check if arrays includes a 3D array
        if 'pos' in arrays:
            arrays.remove('pos')
            for arr in ['x', 'y', 'z']:
                arrays.append(arr)
        if 'vel' in arrays:
            arrays.remove('vel')
            for arr in ['vx', 'vy', 'vz']:
                arrays.append(arr)

        with (
            self.lazy_off,
            util.open_(self.filename, "rb") as fin,
            util.open_(self.filename + ".tmp", "wb") as fout
        ):
            if self._byteswap:
                t, n, ndim, ng, nd, ns = struct.unpack(">diiiii", fin.read(28))
                fout.write(struct.pack(">diiiiii", t, n, ndim, ng, nd, ns, 0))
            else:
                t, n, ndim, ng, nd, ns = struct.unpack("diiiii", fin.read(28))
                fout.write(struct.pack("diiiiii", t, n, ndim, ng, nd, ns, 0))

            fin.read(4)

            if family.gas in fam_out:
                assert(ng == len(self[family.gas]))
            if family.dm in fam_out:
                assert(nd == len(self[family.dm]))
            if family.star in fam_out:
                assert(ns == len(self[family.star]))

            max_block_size = 1024 ** 2  # particles

            # describe the file structure as list of (num_parts,
            # [list_of_properties])
            file_structure = ((ng, family.gas, ["mass", "x", "y", "z", "vx", "vy", "vz", "rho", "temp", "eps", "metals", "phi"]),
                              (nd, family.dm, [
                               "mass", "x", "y", "z", "vx", "vy", "vz", "eps", "phi"]),
                              (ns, family.star, ["mass", "x", "y", "z", "vx", "vy", "vz", "metals", "tform", "eps", "phi"]))

            # do the read/write -- at each block, replace the relevant array

            for n_left, fam, st in file_structure:
                n_done = 0
                self_fam = self[fam]
                while n_left > 0:
                    n_block = min(n_left, max_block_size)

                    # Read in the block
                    if(self._byteswap):
                        g = np.frombuffer(
                            fin.read(len(st) * n_block * 4), 'f').byteswap().reshape((n_block, len(st)))
                    else:
                        g = np.frombuffer(
                            fin.read(len(st) * n_block * 4), 'f').reshape((n_block, len(st)))

                    if fam in fam_out:
                        self_sub = self[fam][n_done:n_done + n_block]
                        # write over the relevant data
                        with self_sub.immediate_mode:
                            for i, name in enumerate(st):

                                if name in arrays:
                                    ar = self_sub[name]
                                    try:
                                        if ar.units != 1 and ar.units != units.NoUnit():
                                            g[:, i] = ar.in_original_units().view(
                                                np.ndarray)
                                        else:
                                            g[:, i] = ar.view(np.ndarray)
                                    except KeyError:
                                        pass

                    # Write out the block
                    if self._byteswap:
                        g.byteswap().tofile(fout)
                    else:
                        g.tofile(fout)

                    # Increment total ptcls read in, decrement ptcls left of
                    # this type
                    n_left -= n_block
                    n_done += n_block

            fin.close()
            fout.close()

            os.system("mv " + self.filename + ".tmp " + self.filename)

    @staticmethod
    def _write(self, filename=None, double_pos=None, double_vel=None, binary_aux_arrays=None, cosmological=True):
        """

        Write a TIPSY (standard) formatted file.

        Additionally, you can specify whether you want position and/or
        velocity arrays written out in double precision. If you are
        writing out a snapshot that was originally in tipsy format and
        the bDoublePos/bDoubleVel flags are set in the parameter file,
        then the write routine will follow those choices. If you are
        writing a snapshot other than a tipsy snapshot, then you have
        to specify these by hand.

        **Optional Keywords**

        *filename* (None): name of the file to be written out. If
                           None, the original file is overwritten.

        *double_pos* (False): set to 'True' if you want to write out positions as doubles

        *double_vel* (False): set to 'True' if you want to write out velocities as doubles

        *binary_aux_arrays* (None): set to 'True' to write auxiliary
                                    arrays in binary format; if left 'None', the preference is
                                    taken from the param file

        *cosmological* (True): if True (default), the header will store the self.properties['a'];
                               otherwise it will store self.properties['time']

        """

        global config

        if filename is None:
            filename = self._filename

        logger.info("Writing main file %s", filename)

        f = util.open_(filename, 'wb')

        t = 0.0
        try:
            if cosmological:
                t = self.properties['a']
            else:
                t = self.properties['time']
        except KeyError:
            warnings.warn(
                "Time is unknown: writing zero in header", RuntimeWarning)

        n = len(self)
        ndim = 3
        ng = len(self.gas)
        nd = len(self.dark)
        ns = len(self.star)
        if n != ng + nd + ns:
            warnings.warn("Snapshot contains extra particles than DM, stars and gas (extra families?). "
                          "Tipsy writing is not set up to handle those. Reverting Tipsy header ntot to ng + nd + ns")
            n = ng + nd + ns

        byteswap = getattr(self, "_byteswap", sys.byteorder == "little")

        if byteswap:
            f.write(struct.pack(">diiiiii", t, n, ndim, ng, nd, ns, 0))
        else:
            f.write(struct.pack("diiiiii", t, n, ndim, ng, nd, ns, 0))

        # needs to be done in blocks like reading
        # describe the file structure as list of (num_parts,
        # [list_of_properties])

        if type(self) is not TipsySnap:
            if double_pos is None:
                double_pos = False
            if double_vel is None:
                double_vel = False
            ptype = 'd' if double_pos else 'f'
            vtype = 'd' if double_vel else 'f'

        else:
            dpos_param = self._paramfile.get('bDoublePos', False)
            dvel_param = self._paramfile.get('bDoubleVel', False)

            if double_pos:
                ptype = 'd'
            elif not double_pos:
                ptype = 'f'
            else:
                ptype = 'd' if dpos_param else 'f'

            if double_vel:
                vtype = 'd'
            elif not double_vel:
                vtype = 'f'
            else:
                vtype = 'd' if dvel_param else 'f'

        g_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "rho", "temp", "eps", "metals", "phi"),
                            'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f', 'f', 'f', 'f')})
        d_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "eps", "phi"),
                            'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f')})
        s_dtype = np.dtype({'names': ("mass", "x", "y", "z", "vx", "vy", "vz", "metals", "tform", "eps", "phi"),
                            'formats': ('f', ptype, ptype, ptype, vtype, vtype, vtype, 'f', 'f', 'f', 'f')})

        file_structure = ((ng, family.gas, g_dtype),
                          (nd, family.dm, d_dtype),
                          (ns, family.star, s_dtype))

        max_block_size = 1024 ** 2  # particles

        with self.lazy_derive_off:
            for n_left, fam, dtype in file_structure:
                n_done = 0
                self_type = self[fam]
                while n_left > 0:
                    n_block = min(n_left, max_block_size)

                #g = np.zeros((n_block,len(st)),dtype=np.float32)

                    g = np.empty(n_block, dtype=dtype)

                    self_type_block = self_type[n_done:n_done + n_block]

                    with self_type_block.immediate_mode:
                        # Copy from the correct arrays
                        for i, name in enumerate(dtype.names):
                            try:
                                g[name] = self_type_block[name]
                            except KeyError:
                                pass

                    # Write out the block
                    if byteswap:
                        g.byteswap().tofile(f)
                    else:
                        g.tofile(f)

                    # Increment total ptcls written, decrement ptcls left of
                    # this type
                    n_left -= n_block
                    n_done += n_block

        f.close()

        logger.info("Writing auxiliary arrays")

        with self.lazy_off:  # prevent any lazy reading or evaluation

            for x in set(self.keys()).union(self.family_keys()):
                if not self.is_derived_array(x) and x not in ["mass", "pos", "x", "y", "z", "vel", "vx", "vy", "vz", "rho", "temp",
                                                              "eps", "metals", "phi", "tform"]:
                    TipsySnap._write_array(
                        self, x, filename=filename + "." + x, binary=binary_aux_arrays)


    @staticmethod
    def __get_write_dtype(stored_dtype):
        if issubclass(stored_dtype.type, np.integer):
            return np.int32
        else:
            return np.float32

    @staticmethod
    def __write_block(f, ar, binary, byteswap):

        ar = np.asarray(ar, dtype=TipsySnap.__get_write_dtype(ar.dtype))
        if binary:
            if byteswap:
                ar.byteswap().tofile(f)
            else:
                ar.tofile(f)

        else:
            if issubclass(ar.dtype.type, np.integer):
                fmt = "%d"
            else:
                fmt = "%e"
            np.savetxt(f, ar, fmt=fmt)

    def _get_loadable_array_metadata(self, array_name):
        """Given an array name, returns the metadata consisting of
        the tuple units, families.

        Returns:
         *units*: the units of this data on disk
         *families*: a list of family objects for which data is on disk
                     for this array, or None if this cannot be determined"""

        try:
            with open(self.filename + "." + array_name + ".pynbody-meta") as f:
                lines = f.readlines()
        except OSError:
            return self._default_units_for(array_name), None, None

        res = {}
        for l in lines:

            X = l.split(":")

            if len(X) == 2:
                res[X[0].strip()] = X[1].strip()

        try:
            u = units.Unit(res['units'])
        except Exception:
            u = None
        try:
            fams = [family.get_family(x.strip())
                    for x in res['families'].split(" ")]
        except Exception:
            fams = None

        try:
            dtype = np.dtype(res['dtype'])
        except Exception:
            dtype = None

        return u, fams, dtype

    @staticmethod
    def _write_array_metafile(self, filename, units, families, dtype):

        with open(filename + ".pynbody-meta", "w") as f:
            print("# This file automatically created by pynbody", file=f)
            if not hasattr(units, "_no_unit"):
                print("units:", units, file=f)
            print("families:", end=' ', file=f)
            for x in families:
                print(x.name, end=' ', file=f)
            print(file=f)
            if dtype is not None:
                print("dtype:", TipsySnap.__get_write_dtype(dtype), file=f)

        if isinstance(self, TipsySnap):
            # update the loadable keys if this operation is likely to have
            # changed them
            self._update_loadable_keys()

    @staticmethod
    def _families_in_main_file(array_name, fam=None):
        fam_for_default = [fX for fX, ars in TipsySnap._basic_loadable_keys.items(
        ) if array_name in ars and fX in fam]
        return fam_for_default

    def _update_array(self, array_name, fam=None,
                      filename=None, binary=None, byteswap=None):
        assert fam is not None

        if self.partial_load:
            raise RuntimeError("Writing back to partially loaded files not yet supported")

        fam_in_main = self._families_in_main_file(array_name, fam)
        if len(fam_in_main) > 0:
            self._update_snapshot(array_name, fam_out=fam_in_main)
            fam = list(set(fam).difference(fam_in_main))
            if len(fam) == 0:
                return

        # If we have disk units for this array, check we can convert into them

        aux_u, aux_f, aux_type = self._get_loadable_array_metadata(array_name)
        if aux_f is None:
            aux_f = self.families()

        if aux_u is not None:
            for f in fam:
                if array_name in self[f]:
                    # check convertible
                    try:
                        self[f][array_name].units.in_units(aux_u)
                    except units.UnitsException:
                        raise OSError(
                            "Units must match the existing auxiliary array on disk.")

        try:
            data = self.__read_array_from_disk(array_name, filename=filename)
        except OSError:
            # doesn't really exist, probably because the other data on disk was
            # in the main snapshot
            self._write_array(self, array_name, fam, filename=filename, binary=binary,
                              byteswap=byteswap)
            return

        for f in fam:
            if aux_u is not None:
                data[self._get_family_slice(f)] = self[f][
                    array_name].in_units(aux_u)
            else:
                data[self._get_family_slice(f)] = self[f][array_name]

        fam = list(set(aux_f).union(fam))

        self._write_array(self, array_name, fam=fam, contents=data, binary=binary,
                          byteswap=byteswap, filename=filename)

    @staticmethod
    def _write_array(self, array_name, fam=None, contents=None,
                     filename=None, binary=None, byteswap=None):
        """Write the array to file."""

        fam_in_main = TipsySnap._families_in_main_file(array_name, fam)
        if len(fam_in_main) > 0:
            if isinstance(self, TipsySnap):
                self._update_snapshot(array_name, fam_out=fam_in_main)
                fam = list(set(fam).difference(fam_in_main))
                if len(fam) == 0:
                    return
            else:
                raise RuntimeError("Cannot call static _write_array to write into main tipsy file.")

        units_out = None
        dtype = None

        if binary is None:
            binary = getattr(self.ancestor, "_tipsy_arrays_binary", False)

        if binary and (byteswap is None):
            byteswap = getattr(self.ancestor, "_byteswap", False)

        with self.lazy_off:  # prevent any lazy reading or evaluation
            if filename is None:
                if self._filename[-3:] == '.gz':
                    filename = self._filename[:-3] + "." + array_name + ".gz"
                else:
                    filename = self._filename + "." + array_name

            if binary:
                fhand = util.open_(filename, 'wb')

                if byteswap:
                    fhand.write(struct.pack(">i", len(self)))

                else:
                    fhand.write(struct.pack("i", len(self)))

            else:
                fhand = util.open_(filename, 'wb')
                fhand.write((str(len(self)) + '\n').encode('utf-8'))

            if contents is None:
                if array_name in self.family_keys():
                    for f in [family.gas, family.dm, family.star]:
                        try:
                            dtype = self[f][array_name].dtype
                            ar = self[f][array_name]
                            units_out = ar.units

                        except KeyError:
                            ar = np.zeros(len(self[f]), dtype=int)

                        TipsySnap.__write_block(fhand, ar, binary, byteswap)

                else:
                    ar = self[array_name]
                    dtype = self[array_name].dtype
                    units_out = ar.units
                    TipsySnap.__write_block(fhand, ar, binary, byteswap)

            else:
                TipsySnap.__write_block(fhand, contents, binary, byteswap)
                units_out = contents.units

        fhand.close()

        if fam is None:
            fam = [family.gas, family.dm, family.star]

        TipsySnap._write_array_metafile(self, filename, units_out, fam, dtype)

    def _load_array(self, array_name, fam=None, filename=None,
                    packed_vector=None):

        if array_name in self._basic_loadable_keys[fam]:
            self._load_main_file()
            return

        fams = self._get_loadable_array_metadata(
            array_name)[1] or self.families()

        if (fam is None and fams is not None and len(fams) != len(self.families())) or \
                (fam is not None and fam not in fams):
            # Top line says 'you requested all families but at least one isn't available'
            # Bottom line says 'you requested one family, but that one's not
            # available'

            raise OSError("This array is marked as available only for families %s" % fams)

        data = self.__read_array_from_disk(array_name, fam=fam,
                                           filename=filename,
                                           packed_vector=packed_vector)

        if fam is None:
            self[array_name] = data
        else:
            self[fam][array_name] = data

    def __read_array_from_disk(self, array_name, fam=None, filename=None,
                               packed_vector=None):
        """Read a TIPSY-ASCII or TIPSY-BINARY auxiliary file with the
        specified name. If fam is not None, read only the particles of
        the specified family."""

        starlog_keys = ['rhoform', 'tempform', 'phiform', 'nsmooth',
                         'xform', 'yform', 'zform', 'vxform', 'vyform', 'vzform',
                         'posform', 'velform','h2form']
        # if massform available as auxiliary array, faster to load it instead
        if array_name == 'massform' and 'massform' not in self.loadable_keys():
            starlog_keys += ['massform']

        if filename is None and array_name in starlog_keys:

            try:
                self.read_starlog()
                if fam is not None:
                    return self[fam][array_name]
                else:
                    return self[array_name]
            except OSError:
                pass

        # N.B. this code is a bit inefficient for loading
        # family-specific arrays, because it reads in the whole array
        # then slices it.  But of course the memory is only wasted
        # while still inside this routine, so it's only really the
        # wasted time that's a problem.

        # determine whether the array exists in a file

        if filename is None:
            if self._filename[-3:] == '.gz':
                filename = self._filename[:-3] + "." + array_name
            else:
                filename = self._filename + "." + array_name


        logger.info("Attempting to load auxiliary array %s", filename)
        # if we get here, we've got the file - try loading it

        units, _, dtype = self._get_loadable_array_metadata(array_name)
        if dtype is None:
            dtype = self._get_preferred_dtype(array_name)

        try:
            f = util.open_(filename, 'r')
            l = int(f.readline())
            binary = False
            if l != self._load_control.disk_num_particles:
                raise OSError("Incorrect file format")

            if not dtype:
                # Inspect the first line to see whether it's float or int
                l = "0\n"
                while l == "0\n":
                    l = f.readline()
                if "." in l or "e" in l or l[:-1] == "inf":
                    dtype = float
                else:
                    dtype = int

                # Restart at head of file
                f.seek(0)
                f.readline()

            loadblock = lambda count: np.fromfile(
                f, dtype=dtype, sep="\n", count=count)
            # data = np.fromfile(f, dtype=tp, sep="\n")
        except ValueError:
            f.close()
            # this is probably a binary file
            binary = True
            f = util.open_(filename, 'rb')
            # Read header and check endianness
            if self._byteswap:
                l = struct.unpack(">i", f.read(4))[0]
            else:
                l = struct.unpack("i", f.read(4))[0]

            if l != self._load_control.disk_num_particles:
                raise OSError("Incorrect file format")

            if dtype is None:
                # Set data format to be read (float or int) based on config
                int_arrays = list(map(
                    str.strip, config_parser.get('tipsy', 'binary-int-arrays').split(",")))
                if array_name in int_arrays:
                    dtype = 'i'
                else:
                    dtype = 'f'

            # Read longest data array possible.
            # Assume byteswap since most will be.
            if self._byteswap:
                loadblock = lambda count: np.frombuffer(
                    f.read(count * 4), dtype=dtype, count=count).byteswap()
                # data = np.fromstring(f.read(3*len(self)*4),dtype).byteswap()
            else:
                loadblock = lambda count: np.frombuffer(
                    f.read(count * 4), dtype=dtype, count=count)
                # data = np.fromstring(f.read(3*len(self)*4),dtype)

        self.ancestor._tipsy_arrays_binary = binary

        all_fam = [family.dm, family.gas, family.star]
        if fam is None:
            fam = all_fam
            r = np.empty(len(self), dtype=dtype).view(array.SimArray)
        else:
            r = np.empty(len(self[fam]), dtype=dtype).view(array.SimArray)

        try:
            for readlen, buf_index, mem_index in self._load_control.iterate(all_fam, fam):
                buf = loadblock(readlen)
                if mem_index is not None:
                    r[mem_index] = buf[buf_index]
        finally:
            f.close()

        if units is not None:
            r.units = units

        return r

    def read_starlog(self, fam=None):
        """Read a TIPSY-starlog file."""

        import pynbody.bridge
        x = os.path.abspath(self._filename)
        done = False
        filename = None
        x = os.path.dirname(x)
        # Attempt the loading of information
        l = glob.glob(os.path.join(x, "*.starlog"))
        if (len(l)):
            for filename in l:
                sl = StarLog(filename)
        else:
            l = glob.glob(os.path.join(x, "../*.starlog"))
            if (len(l) == 0):
                raise OSError("Couldn't find starlog file")
            for filename in l:
                sl = StarLog(filename)

        logger.info("Bridging starlog into SimSnap")
        b = pynbody.bridge.OrderBridge(self, sl)
        b(b(b(sl))).star['iorderGas'] = b(b(sl)).star['iorderGas']
        b(b(b(sl))).star['massform'] = b(b(sl)).star['massform']
        b(b(b(sl))).star['rhoform'] = b(b(sl)).star['rhoform']
        b(b(b(sl))).star['tempform'] = b(b(sl)).star['tempform']
        b(b(b(sl)))['posform'] = b(b(sl))['pos']
        b(b(b(sl)))['velform'] = b(b(sl))['vel']
        #b(sl).star['iorderGas'] = sl.star['iorderGas'][:len(self.star)]
        #b(sl).star['massform'] = sl.star['massform'][:len(self.star)]
        #b(sl).star['rhoform'] = sl.star['rhoform'][:len(self.star)]
        #b(sl).star['tempform'] = sl.star['tempform'][:len(self.star)]
        #b(sl)['posform'] = sl['pos'][:len(self.star), :]
        #b(sl)['velform'] = sl['vel'][:len(self.star), :]
        if 'h2form' in list(sl.star.keys()):
            b(b(b(sl))).star['h2form'] = b(b(sl)).star['h2form']
        else: print("No H2 data found in StarLog file")
        for i, x in enumerate(['x', 'y', 'z']):
            self._arrays[x + 'form'] = self['posform'][:, i]
        for i, x in enumerate(['vx', 'vy', 'vz']):
            self._arrays[x + 'form'] = self['velform'][:, i]

    @classmethod
    def _can_load(cls, f):
        try:
            check = TipsySnap(f, verbose=False)
            del check
        except Exception:
            return False

        return True

# caculate the number fraction YH, YHe as a function of metalicity. Cosmic
# production rate of helium relative to metals (in mass)
# delta Y/delta Z = 2.1 and primordial He Yp = 0.236 (Jimenez et al. 2003,
# Science 299, 5612.
#  piecewise linear
#  Y = Yp + dY/dZ*ZMetal up to ZMetal = 0.1, then linear decrease to 0 at Z=1)

#  SUM_Metal = sum(ni/nH *mi),it is a fixed number for cloudy abundance.
#  Massfraction fmetal = Z*SUM_metal/(1 + 4*nHe/nH + Z*SUM_metal) (1)
#  4*nHe/nH = mHe/mH = fHe/fH
#  also fH + fHe + fMetal = 1  (2)
#  if fHe specified, combining the 2 eq above will solve for
#  fH and fMetal


def _abundance_estimator(metal):

    Y_He = ((0.236 + 2.1 * metal) / 4.0) * (metal <= 0.1)
    Y_He += ((-0.446 * (metal - 0.1) / 0.9 + 0.446) / 4.0) * (metal > 0.1)
    Y_H = 1.0 - Y_He * 4. - metal

    return Y_H, Y_He


@TipsySnap.derived_quantity
def HII(sim):
    """Number of HII ions per proton mass"""
    Y_H, Y_He = _abundance_estimator(sim["metals"])
    try:
        return Y_H - sim["HI"] - 2 * sim["H2"]
    except KeyError:
        return Y_H - sim["HI"]


@TipsySnap.derived_quantity
def HeIII(sim):
    """Number of HeIII ions per proton mass"""
    Y_H, Y_He = _abundance_estimator(sim["metals"])
    return Y_He - sim["HeII"] - sim["HeI"]


@TipsySnap.derived_quantity
def ne(sim):
    """Number of electrons per proton mass"""
    return sim["HII"] + sim["HeII"] + 2 * sim["HeIII"]


@TipsySnap.derived_quantity
def hetot(self):
    """Helium mass fraction including correction based on metallicity"""
    return 0.236 + (2.1 * self['metals'])


@TipsySnap.derived_quantity
def hydrogen(self):
    """Hydrogen mass fraction including correction based on metallicity"""
    return 1.0 - self['metals'] - self['hetot']

#from .tipsy import TipsySnap
# Asplund et al (2009) ARA&A solar abundances (Table 1)
# m_frac = 10.0^([X/H] - 12)*M_X/M_H*0.74
# OR
# http://en.wikipedia.org/wiki/Abundance_of_the_chemical_elements
# puts stuff more straighforwardly cites Arnett (1996)
# A+G from http://www.t4.lanl.gov/opacity/grevand1.html
# Anders + Grev (1989)    Asplund
XSOLFe = 0.125E-2         # 1.31e-3
# Looks very wrong ([O/Fe] ~ 0.2-0.3 higher than solar),
# probably because SN ejecta are calculated with
# Woosley + Weaver (1995) based on Anders + Grevesse (1989)
# XSOLO=0.59E-2           # 5.8e-2
XSOLO = 0.84E-2
XSOLH = 0.706             # 0.74
XSOLC = 3.03e-3           # 2.39e-3
XSOLN = 9.2e-4          # 7e-4
XSOLNe = 1.66e-3          # 1.26e-3
XSOLMg = 6.44e-4          # 7e-4
XSOLSi = 7e-4          # 6.7e-4


@TipsySnap.derived_quantity
def feh(self):
    """Iron abundance [Fe/H] derived from tipsy array FeMassFrac, with solar
    values from Asplund et al 09"""
    minfe = np.amin(self['FeMassFrac'][np.where(self['FeMassFrac'] > 0)])
    self['FeMassFrac'][np.where(self['FeMassFrac'] == 0)] = minfe
    return np.log10(self['FeMassFrac'] / self['hydrogen']) - np.log10(XSOLFe / XSOLH)


@TipsySnap.derived_quantity
def oxh(self):
    """Oxygen abundance [O/H] derived from tipsy array FeMassFrac, with solar
    values from Asplund et al 09"""
    minox = np.amin(self['OxMassFrac'][np.where(self['OxMassFrac'] > 0)])
    self['OxMassFrac'][np.where(self['OxMassFrac'] == 0)] = minox
    return np.log10(self['OxMassFrac'] / self['hydrogen']) - np.log10(XSOLO / XSOLH)


@TipsySnap.derived_quantity
def ofe(self):
    """Oxygen-to-iron ratio [O/Fe] derived from tipsy arrays OxMassFrac and FeMassFrac
    with solar values from Asplund et al 09"""
    minox = np.amin(self['OxMassFrac'][np.where(self['OxMassFrac'] > 0)])
    self['OxMassFrac'][np.where(self['OxMassFrac'] == 0)] = minox
    minfe = np.amin(self['FeMassFrac'][np.where(self['FeMassFrac'] > 0)])
    self['FeMassFrac'][np.where(self['FeMassFrac'] == 0)] = minfe
    return np.log10(self['OxMassFrac'] / self['FeMassFrac']) - np.log10(XSOLO / XSOLFe)


@TipsySnap.derived_quantity
def mgfe(sim):
    """Magnesium-to-iron ratio [Mg/Fe] derived from tipsy arrays MgMassFrac and FeMassFrac
    with solar values from Asplund et al 09"""
    minmg = np.amin(sim['MgMassFrac'][np.where(sim['MgMassFrac'] > 0)])
    sim['MgMassFrac'][np.where(sim['MgMassFrac'] == 0)] = minmg
    minfe = np.amin(sim['FeMassFrac'][np.where(sim['FeMassFrac'] > 0)])
    sim['FeMassFrac'][np.where(sim['FeMassFrac'] == 0)] = minfe
    return np.log10(sim['MgMassFrac'] / sim['FeMassFrac']) - np.log10(XSOLMg / XSOLFe)


@TipsySnap.derived_quantity
def nefe(sim):
    """Neon-to-iron ratio [Ne/Fe] derived from tipsy arrays MgMassFrac and FeMassFrac
    with solar values from Asplund et al 09"""
    minne = np.amin(sim['NeMassFrac'][np.where(sim['NeMassFrac'] > 0)])
    sim['NeMassFrac'][np.where(sim['NeMassFrac'] == 0)] = minne
    minfe = np.amin(sim['FeMassFrac'][np.where(sim['FeMassFrac'] > 0)])
    sim['FeMassFrac'][np.where(sim['FeMassFrac'] == 0)] = minfe
    return np.log10(sim['NeMassFrac'] / sim['FeMassFrac']) - np.log10(XSOLNe / XSOLFe)


@TipsySnap.derived_quantity
def sife(sim):
    """Silicon-to-iron ratio [Si/Fe] derived from tipsy arrays MgMassFrac and FeMassFrac
    with solar values from Asplund et al 09"""
    minsi = np.amin(sim['SiMassFrac'][np.where(sim['SiMassFrac'] > 0)])
    sim['SiMassFrac'][np.where(sim['SiMassFrac'] == 0)] = minsi
    minfe = np.amin(sim['FeMassFrac'][np.where(sim['FeMassFrac'] > 0)])
    sim['FeMassFrac'][np.where(sim['FeMassFrac'] == 0)] = minfe
    return np.log10(sim['SiMassFrac'] / sim['FeMassFrac']) - np.log10(XSOLSi / XSOLFe)


@TipsySnap.derived_quantity
def c_s(self):
    """Ideal gas sound speed based on pressure and density"""
    #x = np.sqrt(5./3.*units.k*self['temp']*self['mu'])
    x = np.sqrt(5. / 3. * self['p'] / self['rho'])
    x.convert_units('km s^-1')
    return x


@TipsySnap.derived_quantity
def c_s_turb(self):
    """Turbulent sound speed (from Mac Low & Klessen 2004)"""
    x = np.sqrt(self['c_s'] ** 2 + 1. / 3 * self['v_disp'] ** 2)
    x.convert_units('km s^-1')
    return x


@TipsySnap.derived_quantity
def mjeans(self):
    """Classical Jeans mass"""
    x = np.pi ** (5. / 2.) * self['c_s'] ** 3 / \
        (6. * units.G ** (3, 2) * self['rho'] ** (1, 2))
    x.convert_units('Msol')
    return x


@TipsySnap.derived_quantity
def mjeans_turb(self):
    """Turbulent Jeans mass"""
    x = np.pi ** (5. / 2.) * self['c_s_turb'] ** 3 / \
        (6. * units.G ** (3, 2) * self['rho'] ** (1, 2))
    x.convert_units('Msol')
    return x


@TipsySnap.derived_quantity
def ljeans(self):
    """Jeans length"""
    x = self['c_s'] * np.sqrt(np.pi / (units.G * self['rho']))
    x.convert_units('kpc')
    return x


@TipsySnap.derived_quantity
def ljeans_turb(self):
    """Turbulent Jeans length"""
    x = self['c_s_turb'] * np.sqrt(np.pi / (units.G * self['rho']))
    x.convert_units('kpc')
    return x


class StarLog(SimSnap):

    def __init__(self, filename, sort=True, paramfile=None, use_log=True):
        import os
        super().__init__()
        self._filename = filename
        self._paramfilename = paramfile

        f = util.open_(filename, "rb")
        self.properties = {}

        size = struct.unpack("i", f.read(4))
        if (size[0] > 1000 or size[0] < 10):
            self._byteswap = True
            f.seek(0)
            size = struct.unpack(">i", f.read(4))
        iSize = size[0]

        bigstarlog = False
        molecH = False
        bigIOrds = False

        # replace only the .starlog suffix with .log
        self._logfile = '.'.join([x for x in filename.split('.')[:-1]]) + '.log'

        if use_log:
            # assumes log file would be in the same location as starlog file
            logger.info('Attempting to load starlog metadata from log file')
            try:
                with open(self._logfile) as g:
                    read_metadata = False
                    structure_names = []
                    structure_formats = []
                    for line in g:
                        if line.startswith('# end starlog data'):
                            break
                        if read_metadata:
                            meta_name, meta_type = line.strip().strip('#').split()
                            meta_name = self._infer_name_from_tipsy_log(meta_name)
                            structure_names.append(meta_name)
                            structure_formats.append(meta_type)
                        if line.startswith('# starlog data:'):
                            read_metadata = True
                file_structure = np.dtype({'names': structure_names,
                                           'formats': structure_formats})
                if file_structure.itemsize != iSize:
                    raise ValueError('Starlog metadata size does not match with starlog file size')
                if file_structure['iord'] == 'f8':
                	bigIOrds = True
                if 'h2form' in file_structure.names:
                	molecH = True
            except FileNotFoundError:
                warnings.warn('No log file found; reverting to guess-and-check')
                use_log = False
            except ValueError:
                warnings.warn('log file found, but there was a problem with the starlog metadata. '
                              'Reverting to guess-and-check')
                use_log = False




        if use_log is False:
            file_structure = np.dtype({'names': ("iord", "iorderGas", "tform",
                                                 "x", "y", "z",
                                                 "vx", "vy", "vz",
                                                 "massform", "rhoform", "tempform"),
                                       'formats': ('i4', 'i4', 'f8',
                                                   'f8', 'f8', 'f8',
                                                   'f8', 'f8', 'f8',
                                                   'f8', 'f8', 'f8')})

            if (iSize > file_structure.itemsize):
                file_structure = np.dtype({'names': ("iord", "iorderGas", "tform",
                                                 "x", "y", "z",
                                                 "vx", "vy", "vz",
                                                 "massform", "rhoform", "tempform","h2form"),
                                       'formats': ('i4', 'i4', 'f8',
                                                   'f8', 'f8', 'f8',
                                                   'f8', 'f8', 'f8',
                                                   'f8', 'f8', 'f8','f8')})
                molecH = True
                # Unfortunately molecularH with small iOrders has the same as
                # no moleculuarH with big iOrders.  Attempt to distinguish here
                if(iSize == file_structure.itemsize):
                    if(self._byteswap):
                        testread = np.frombuffer(
                            f.read(iSize), dtype=file_structure).byteswap()
                    else:
                        testread = np.frombuffer(f.read(iSize), dtype=file_structure)
                    # All star iorders are greater than any gas iorder
                    # so this indicates a bad format. (N.B. there is the
                    # possibility of a false negative)
                    if(testread['iord'][0] < testread['iorderGas'][0]):
                        file_structure = np.dtype({'names': ("iord", "iorderGas",
                                                 "tform",
                                                 "x", "y", "z",
                                                 "vx", "vy", "vz",
                                                 "massform", "rhoform", "tempform"),
                                           'formats': ('i8', 'i8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8')})
                        f.seek(4)
                        logger.info("Using 64 bit iOrders")
                        molecH = False
                        bigIOrds = True
            if (iSize != file_structure.itemsize):
                file_structure = np.dtype({'names': ("iord", "iorderGas", "tform",
                                                     "x", "y", "z",
                                                     "vx", "vy", "vz",
                                                     "massform", "rhoform", "tempform",
                                                     "phiform", "nsmooth"),
                                           'formats': ('i4', 'i4', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'i4')})
                molecH = False

                if (iSize != file_structure.itemsize and iSize != 104):
                    raise OSError("Unknown starlog structure iSize:" + \
                        str(iSize) + ", file_structure itemsize:" + \
                        str(file_structure.itemsize))
                else:
                    bigstarlog = True
            if molecH is True: print("h2 information found in StarLog!")

        datasize = os.path.getsize(filename) - f.tell()

        # check whether datasize is a multiple of iSize. If it is not,
        # the starlog is likely corrupted, but try to read it anyway

        if (datasize % iSize > 0) and (iSize != 104):
            warnings.warn(
                "The size of the starlog file does not make sense -- it is likely corrupted. Pynbody will read it anyway, but use with caution.")
            datasize -= datasize % iSize

        logger.info("Reading starlog file %s", filename)
        if(self._byteswap):
            g = np.frombuffer(
                f.read(datasize), dtype=file_structure).byteswap()
        else:
            g = np.frombuffer(f.read(datasize), dtype=file_structure)

        f.close()
        # hoping to provide backward compatibility for np.unique by
        # copying relavent part of current numpy source:
        # numpy/lib/arraysetops.py:192 (on 22nd March 2011)

        if sort:
            tmp = g['iord'].flatten()
            # mergesort for stability
            perm = tmp.argsort(kind='mergesort')
            aux = tmp[perm]
            flag = np.concatenate((aux[1:] != aux[:-1], [True]))
            iord = aux[flag]
            indices = perm[flag]
            self._num_particles = len(indices)
        else:
            self._num_particles = len(g)

        self._family_slice[family.star] = slice(0, self._num_particles)
        self._create_arrays(["pos", "vel"], 3)
        if(bigIOrds):
            self._create_arrays(["iord"], dtype='int64')
        else:
            self._create_arrays(["iord"], dtype='int32')
        self._create_arrays(["iord"], dtype='int32')
        self._create_arrays(
            ["iorderGas", "massform", "rhoform", "tempform", "metals", "tform"])
        if molecH:
            self._create_arrays(["h2form"])
        if bigstarlog:
            self._create_arrays(["phiform", "nsmooth"])

        self._decorate()

        if sort:
            for name in list(file_structure.fields.keys()):
                self.star[name][:] = g[name][indices]
        else:
            for name in list(file_structure.fields.keys()):
                self.star[name] = g[name]

    @staticmethod
    def _write(self, filename=None):
        """Write the starlog file. """

        global config

        with self.lazy_off:

            if filename is None:
                filename = self._filename

            logger.info("Writing starlog file as %s", filename)

            f = util.open_(filename, 'wb')

            if 'phiform' in list(self.keys()):  # long starlog format
                file_structure = np.dtype({'names': ("iord", "iorderGas", "tform",
                                                     "x", "y", "z",
                                                     "vx", "vy", "vz",
                                                     "massform", "rhoform", "tempform",
                                                     "phiform", "nsmooth"),
                                           'formats': ('i4', 'i4', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'i4')})
            else:  # short (old) starlog format
                file_structure = np.dtype({'names': ("iord", "iorderGas", "tform",
                                                     "x", "y", "z",
                                                     "vx", "vy", "vz",
                                                     "massform", "rhoform", "tempform"),
                                           'formats': ('i4', 'i4', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8',
                                                       'f8', 'f8', 'f8')})

            if self._byteswap:
                f.write(struct.pack(">i", file_structure.itemsize))
            else:
                f.write(struct.pack("i", file_structure.itemsize))

            max_block_size = 1024  # particles

            n_left = len(self)
            n_done = 0

            while n_left > 0:
                n_block = min(n_left, max_block_size)

                g = np.zeros(n_block, dtype=file_structure)

                for arr in file_structure.names:
                    g[arr] = self[arr][n_done:n_done + n_block]

                if self._byteswap:
                    g.byteswap().tofile(f)
                else:
                    g.tofile(f)

                n_left -= n_block
                n_done += n_block

        f.close()

    def _infer_name_from_tipsy_log(self, meta_name):
        """Convert starlog metadata name to pynbody array name

        In the future, it may be necessary to expand this or make
        it more flexible. For now, a dictionary should do.
        """
        conversion_dict = {'iOrdStar': 'iord', 'iOrdGas': 'iorderGas',
                           'timeForm': 'tform',
                           'rForm[0]': 'x', 'rForm[1]': 'y', 'rForm[2]': 'z',
                           'vForm[0]': 'vx', 'vForm[1]': 'vy', 'vForm[2]': 'vz',
                           'massForm': 'massform', 'rhoForm': 'rhoform',
                           'TForm': 'tempform', 'tCoolForm': 'tcoolform',
                           'H2FracForm': 'h2form'}
        if meta_name in conversion_dict.keys():
            return conversion_dict[meta_name]
        else:  # this is where cleverness will be needed in the future
            raise ValueError('Unknown starlog entry: ' + meta_name)


@TipsySnap.decorator
@StarLog.decorator
@nchilada.NchiladaSnap.decorator
def load_paramfile(sim):
    x = os.path.dirname(os.path.abspath(sim._filename))
    sim._paramfile = {}
    ok = False
    if sim._paramfilename is None:
        candidates = [
            *glob.glob(os.path.join(x, "*.param")),
            *glob.glob(os.path.join(x, "../*.param")),
        ]
    else:
        candidates = [sim._paramfilename]

    for filename in candidates:
        try:
            with open(filename):
                pass
            ok = True
            break
        except OSError:
            pass


    if not ok:
        if sim._paramfile:
            raise OSError("The parameter filename you supplied is invalid")
        return

    with open(filename) as f:
        lines = f.readlines()

    for line in lines:
        if line.startswith("#"):
            continue
        # Remove comments
        line = line.split("#")[0]

        # Lines are "key  =  val"
        try:
            key, val = (_.strip() for _ in line.split("="))
        except ValueError:
            continue
        sim._paramfile[key] = val

    if len(sim._paramfile) > 1:
        sim._paramfile["filename"] = filename

@TipsySnap.decorator
@StarLog.decorator
@nchilada.NchiladaSnap.decorator
def param2units(sim):
    with sim.lazy_off:
        munit = dunit = hub = None

        logger.debug("paramfile: %s", sim._paramfile)

        try:
            hub = float(sim._paramfile["dHubble0"])
            sim.properties['omegaM0'] = float(sim._paramfile["dOmega0"])
            sim.properties['omegaL0'] = float(sim._paramfile["dLambda"])
        except KeyError:
            pass

        try:
            munit_st = sim._paramfile["dMsolUnit"] + " Msol"
            munit = float(sim._paramfile["dMsolUnit"])
            dunit_st = sim._paramfile["dKpcUnit"] + " kpc"
            dunit = float(sim._paramfile["dKpcUnit"])
            logger.debug("munit: %s, dunit: %s", munit_st, dunit_st)
        except KeyError:
            pass

        if munit is None or dunit is None:
            if hub is not None:
                sim.properties['h'] = hub
            return

        denunit = munit / dunit ** 3
        denunit_st = str(denunit) + " Msol kpc^-3"

        #
        # the obvious way:
        #
        #denunit_cgs = denunit * 6.7696e-32
        #kpc_in_km = 3.0857e16
        #secunit = 1./math.sqrt(denunit_cgs*6.67e-8)
        #velunit = dunit*kpc_in_km/secunit

        # the sensible way:
        # avoid numerical accuracy problems by concatinating factors:
        velunit = 8.0285 * math.sqrt(6.6743e-8 * denunit) * dunit
        velunit_st = ("%.5g" % velunit) + " km s^-1"

        # You have: kpc s / km
        # You want: Gyr
        #* 0.97781311
        timeunit = dunit / velunit * 0.97781311
        timeunit_st = ("%.5g" % timeunit) + " Gyr"

        #  Assuming G=1 in code units, phi is actually vel^2/a^3.
        # See also gasoline's master.c:5511.
        # Or should we be calculating phi as GM/R units (which
        # is the same for G=1 runs)?
        potunit_st = "%.5g km^2 s^-2" % (velunit ** 2)

        if 'bComove' in sim._paramfile and int(sim._paramfile['bComove']) != 0:
            hubunit = 10. * velunit / dunit
            hubunit_st = ("%.3f" % (hubunit * hub))
            sim.properties['h'] = hub * hubunit

            if isinstance(sim, StarLog):
                a = "aform"
            else:
                a = "a"

            # append dependence on 'a' for cosmological runs
            dunit_st += " " + a
            denunit_st += " " + a + "^-3"
            velunit_st += " " + a
            potunit_st += " " + a + "^-1"

            # Assume the box size is equal to the length unit
            sim.properties['boxsize'] = units.Unit(dunit_st)

        try:
            sim["vel"].units = velunit_st
        except KeyError:
            pass

        try:
            sim["phi"].units = potunit_st
            sim["eps"].units = dunit_st
        except KeyError:
            pass

        try:
            sim["pos"].units = dunit_st
        except KeyError:
            pass

        try:
            sim.gas["rho"].units = denunit_st
        except KeyError:
            pass

        try:
            sim["mass"].units = munit_st
        except KeyError:
            pass

        try:
            sim.star["tform"].units = timeunit_st
        except KeyError:
            pass

        try:
            sim.gas["metals"].units = ""
        except KeyError:
            pass

        try:
            sim.star["metals"].units = ""
        except KeyError:
            pass

        try:
            sim._file_units_system = [
                sim["vel"].units, sim.star["mass"].units, sim["pos"].units, units.K]
            logger.debug("file_units_system set from arrays %s", sim._file_units_system)
        except KeyError:
            try:
                sim._file_units_system = [
                    sim["vel"].units, sim.star["massform"].units, sim["pos"].units, units.K]
                logger.debug("file_units_system set from arrays inc star massform %s", sim._file_units_system)
            except KeyError:
                try:
                    sim._file_units_system = [units.Unit(velunit_st),
                                              units.Unit(munit_st),
                                              units.Unit(dunit_st), units.K]
                    logger.debug("file_units_system set from paramfile %s", sim._file_units_system)
                except Exception:
                    pass

    sim._override_units_system()


@TipsySnap.decorator
def settime(sim):
    if 'bComove' in sim._paramfile and int(sim._paramfile['bComove']) != 0:
        t = sim._header_t
        sim.properties['a'] = t
        try:
            sim.properties['z'] = 1.0 / t - 1.0
        except ZeroDivisionError:
            # no sensible redshift
            pass

        if (sim.properties['z'] is not None and
                'dMsolUnit' in sim._paramfile and
                'dKpcUnit' in sim._paramfile):
            from ..analysis import cosmology
            sim.properties['time'] = cosmology.age(
                sim, unit=sim.infer_original_units('yr'))
        else:
            # something has gone wrong with the cosmological side of
            # things
            warnings.warn(
                "Paramfile suggests time is cosmological, but header values are not sensible in this context.", RuntimeWarning)
            sim.properties['time'] = t

        sim.properties['a'] = t

    else:
        # Assume a non-cosmological run
        sim.properties['time'] = sim._header_t

@nchilada.NchiladaSnap.decorator
def settimeN(sim):
    if 'bComove' in sim._paramfile and int(sim._paramfile['bComove']) != 0:
        #t = sim._header_t
        #sim.properties['a'] = t
        try:
            sim.properties['z'] = 1.0 / sim.properties['a'] - 1.0
        except ZeroDivisionError:
            # no sensible redshift
            pass

        if (sim.properties['z'] is not None and
                'dMsolUnit' in sim._paramfile and
                'dKpcUnit' in sim._paramfile):
            from ..analysis import cosmology
            sim.properties['time'] = cosmology.age(
                sim, unit=sim.infer_original_units('yr'))
        else:
            # something has gone wrong with the cosmological side of
            # things
            warnings.warn(
                "Paramfile suggests time is cosmological, but header values are not sensible in this context.", RuntimeWarning)
            sim.properties['time'] = sim.properties['a']

        #sim.properties['a'] = t

    else:
        # Assume a non-cosmological run
        sim.properties['time'] = sim.properties['a']

    time_unit = None
    try:
        time_unit = sim.infer_original_units('yr')
    except units.UnitsException:
        pass

    if time_unit is not None:
        sim.properties['time'] *= time_unit


@StarLog.decorator
def slparam2units(sim):
    with sim.lazy_off:
        munit = dunit = hub = None

        try:
            munit_st = sim._paramfile["dMsolUnit"] + " Msol"
            munit = float(sim._paramfile["dMsolUnit"])
            dunit_st = sim._paramfile["dKpcUnit"] + " kpc"
            dunit = float(sim._paramfile["dKpcUnit"])
            hub = float(sim._paramfile["dHubble0"])
        except KeyError:
            pass

        if munit is None or dunit is None:
            return

        denunit = munit / dunit ** 3
        denunit_st = str(denunit) + " Msol kpc^-3"
        velunit = 8.0285 * math.sqrt(6.6743e-8 * denunit) * dunit
        velunit_st = ("%.5g" % velunit) + " km s^-1"

        # You have: kpc s / km
        # You want: Gyr
        #* 0.97781311
        timeunit = dunit / velunit * 0.97781311
        timeunit_st = ("%.5g" % timeunit) + " Gyr"


        if hub is not None:
            # append dependence on 'a' for cosmological runs
            dunit_st += " aform"

            # denunit_st += " a^-3"
            # N.B. density comoving -> physical conversion is done by Gasoline
            # itself

        sim.star["rhoform"].units = denunit_st
        sim.star["massform"].units = munit_st



"""

chunk
=====

Methods for describing parts of files to load.

This module provides generalized logic for getting parts of sequential data off disk.
It is for internal use. If you want to write a loader that supports partial loading,
it will make it a lot easier.

The steps for loading particle data are as follows

1. Set up a description of the particles you have on disk. This is a dictionary
mapping a family type to a slice, e.g.

  on_disk = {pynbody.family.dm: slice(0,100), pynbody.family.gas: slice(100, 150)}

describes a file with 100 dark matter particles followed by 50 gas particles.


2. Create a LoadControl object as defined in this module.

 load_control = pynbody.chunk.LoadControl(on_disk, chunk_size, take)

Here,

 * on_disk is the dictionary you set up in the last step.

 * chunk_size is the maximum number of particles you are willing to load off
  disk at once these will have to be stored in a temporary array, so you don't
  want it to be too large; but also reading a small number of large chunks is
  more efficient, so you don't want it to be too small. No careful experimentation
  has been done with this, but chunk_sizes of around 10000 seem to work OK.

 * take describes what to load in. Currently this is either None (= load the whole file)
   or a list of ids (= load the specified particles). However this may be expanded
   in future to a more comprehensive syntax. The idea is your code will not have to
   change when this happens, and will automatically support more advanced partial loading
   specifications.

3. Load your particle data. The load_control object has an iterate method. This returns
step-by-step instructions that take you through the file, specifying what to keep and what
to throw away in a simple-to-use fashion. See the help for LoadControl.iterate for details
on how to implement this final step.

"""


import copy
import math
import random

import numpy as np

from .. import util


class Chunk:

    def __init__(self, *args, **kwargs):
        pass
        start, stop, step = 0, None, 1
        if len(args) == 1:
            start, stop, step = 0, args[0], 1
        if len(args) == 2:
            start, stop, step = args[0], args[1], 1
        if len(args) == 3:
            start, stop, step = args
        self.random = kwargs.get('random', None)
        self.ids = kwargs.get('ids', None)

        self.start = start
        self.stop = stop
        self.step = step

        assert (self.step is not None
                or self.random is not None
                or self.ids is not None)

    def init(self, max_stop):

        assert max_stop >= 0

        if self.stop is None:
            self.stop = max_stop
        else:
            self.stop = min(max_stop, self.stop)

        assert self.start >= 0
        assert self.stop >= 0, '%ld' % self.stop
        assert self.step > 0
        assert self.start <= self.stop

        if self.random is not None:
            assert random > 0

        if self.ids is not None:
            assert np.amax(self.ids) < max_stop
            if self.random is not None:
                if self.random < 1:
                    self.random = int(self.random * len(self.ids))
                self.ids = random.sample(self.ids, self.random)
                self.ids.sort()
            self.len = len(self.ids)
        else:
            slice_len = int(math.ceil((self.stop - self.start) / self.step))
            self.len = slice_len
            if self.random is not None:
                if self.random < 1:
                    self.random = int(self.random * self.len)
                self.random = min(self.random, slice_len)
                self.ids = random.sample(
                    range(self.start, self.stop, self.step), self.random)
                self.ids.sort()
                self.len = len(self.ids)

    def __len__(self):
        return self.len

    def pdeltas(self, step=None):

        assert step is None or self.contiguous()
        if step is None:
            step = self.step

        if self.ids is None:
            for i in range(self.start, self.stop, self.step):
                yield self.step
        else:
            yield self.ids[0]
            for i in range(len(self.ids) - 1):
                yield self.ids[i + 1] - self.ids[i]

    def contiguous(self):
        return self.ids is None and self.step == 1


class LoadControl:

    """LoadControl provides the logic required for partial loading."""

    def __init__(self, family_slice, max_chunk, clauses):
        """Initialize a LoadControl object.

        *Inputs:*

          *family_slice*: a dictionary of family slices describing the contiguous
            layout of families on disk

          *max_chunk*: the guaranteed maximum chunk of data to load in a single
            read operation. Larger values are likely more efficient, but also require
            bigger temporary buffers in your reader code.

          *clauses*: a dictionary describing the type of partial loading to implement.
            If this dictionary is empty, all data is loaded. Otherwise it can contain
            'ids', a list of ids to load.

         """

        self._disk_family_slice = family_slice
        self._generate_family_order()

        # generate simulation-level ID list
        if hasattr(clauses, "__len__"):
            self._ids = np.asarray(clauses)
        else:
            self._ids = None  # no partial loading!

        self.generate_family_id_lists()
        self._generate_mem_slice()

        self.mem_num_particles = self.mem_family_slice[
            self._ordered_families[-1]].stop
        self.disk_num_particles = self._disk_family_slice[
            self._ordered_families[-1]].stop

        self._generate_chunks(max_chunk)

    @staticmethod
    def _scan_for_next_stop(ids, offset_start, id_maximum):
        from . import scan
        return scan.scan_for_next_stop(ids, offset_start, id_maximum)

    def generate_family_id_lists(self):
        if self._ids is None:
            self._family_ids = None
            return

        self._family_ids = {}
        offset = 0
        stop = 0
        for fam in self._ordered_families:
            sl = self._disk_family_slice[fam]
            self._family_ids[fam] = self._ids[
                (self._ids >= sl.start) * (self._ids < sl.stop)] - sl.start

    def _generate_family_order(self):
        famlist = []
        for fam, sl in self._disk_family_slice.items():
            famlist.append((fam, sl.start, sl.stop))

        famlist.sort(key=lambda x: x[1])
        famlist.sort(key=lambda x: x[2])
        self._ordered_families = [x[0] for x in famlist]

    def _generate_mem_slice(self):
        if self._ids is None:
            self.mem_family_slice = self._disk_family_slice
            return

        self.mem_family_slice = {}
        stop = 0
        for current_family in self._ordered_families:

            start = stop
            stop = stop + len(self._family_ids[current_family])
            self.mem_family_slice[current_family] = slice(start, stop)

    def _generate_null_chunks(self, max_chunk):
        """Generate internal chunk map in the special case that we are loading
        all data"""

        self._family_chunks = {}

        for current_family in self._ordered_families:
            self._family_chunks[current_family] = []
            disk_sl = self._disk_family_slice[current_family]
            for i0 in range(0, disk_sl.stop - disk_sl.start, max_chunk):
                nread = min(disk_sl.stop - disk_sl.start - i0, max_chunk)
                buf_sl = slice(0, nread)
                mem_sl = slice(i0, i0 + nread)

                self._family_chunks[current_family].append(
                    (nread, buf_sl, mem_sl))

    def _generate_chunks(self, max_chunk):
        """Generate internal chunk map, with maximum load chunk of specified number
        of entries, and with chunks that do not cross family boundaries."""

        if self._ids is None:
            self._generate_null_chunks(max_chunk)
            return

        self._family_chunks = {}

        for current_family in self._ordered_families:
            self._family_chunks[current_family] = []
            disk_sl = self._disk_family_slice[current_family]
            ids = self._family_ids[current_family]
            i = 0

            disk_ptr = 0
            mem_ptr = 0
            # print current_family, disk_sl.stop

            while disk_ptr < disk_sl.stop - disk_sl.start:
                disk_ptr_end = disk_ptr + \
                    min(disk_sl.stop - disk_sl.start - disk_ptr, max_chunk - 1)
                j = self._scan_for_next_stop(ids, i, disk_ptr_end - 1)

                nread_disk = disk_ptr_end - disk_ptr

                assert (ids[i:j] < disk_ptr_end).all()

                if i != j:
                    mem_slice = slice(mem_ptr, mem_ptr + j - i)
                else:
                    mem_slice = None

                disk_mask = ids[i:j] - disk_ptr

                # print mem_slice, (j-i), len(disk_mask), disk_ptr,
                # disk_ptr_end, nread_disk

                mem_ptr = mem_ptr + j - i
                i = j
                disk_ptr = disk_ptr_end

                self._family_chunks[current_family].append(
                    (nread_disk, disk_mask, mem_slice))

    def iterate_with_interrupts(self, families_on_disk, families_in_memory,
                                disk_interrupt_points, disk_interrupt_fn, multiskip=False):
        """Performs the same function as
        :func:`~pynbody.chunk.LoadControl.iterate` but additionally
        takes a list of exact file offsets *disk_interrupt_points* at
        which to interrupt the loading process and call a
        user-specified function *disk_interrupt_fn*.

        **Input:** :
          *disk_interrupt_points*: a list (or other iterable) of disk offsets
            (must be in ascending order) at which to call the interrupt function
          *disk_interrupt_fn*: a function which takes the file offset as an argument,
            and is called precisely at the point that the disk interrupt point is reached

          See func:`~pynbody.chunk.LoadControl.iterate` for other arguments.

          """

        fpos = 0
        i = 0
        next_dip = disk_interrupt_points[i]
        for nread_disk, disk_slice, mem_slice in self.iterate(families_on_disk, families_in_memory, multiskip):
            while next_dip and fpos + nread_disk > next_dip:

                # print "ORIGINAL:",nread_disk, disk_slice, mem_slice,"starts@",fpos
                # print "   INTERRUPT AT:",next_dip

                # an interrupt falls in the middle of our slice
                # work out what to read first
                len_pre = next_dip - fpos

                if disk_slice is not None:
                    d_slice_pre = util.concatenate_indexing(
                        slice(0, len_pre), disk_slice)
                    len_m_pre = util.indexing_length(d_slice_pre)
                    m_slice_pre = util.concatenate_indexing(
                        mem_slice, slice(0, len_m_pre))

                    # work out what to read second
                    len_post = nread_disk - len_pre
                    d_slice_post = copy.copy(util.concatenate_indexing(
                        disk_slice, slice(len_m_pre, util.indexing_length(mem_slice))))
                    # the copy above is necessary to ensure we don't end up inadvertently modifying
                    # list of offsets somewhere else

                    # that's the disk slice relative to having read the whole thing continuously.
                    # Offset to reflect what we've missed.

                    if isinstance(d_slice_post, slice):
                        d_slice_post = slice(
                            d_slice_post.start - len_pre, d_slice_post.stop - len_pre, d_slice_post.step)
                    else:
                        d_slice_post -= len_pre

                    len_m_post = util.indexing_length(d_slice_post)
                    m_slice_post = util.concatenate_indexing(
                        mem_slice, slice(len_m_pre, len_m_pre + len_m_post))

                    if util.indexing_length(d_slice_post) == 0:
                        d_slice_post = None
                        m_slice_post = None

                    if util.indexing_length(d_slice_pre) == 0:
                        d_slice_pre = None
                        m_slice_pre = None

                else:
                    d_slice_pre = None
                    m_slice_pre = None
                    len_post = nread_disk - len_pre
                    d_slice_post = None
                    m_slice_post = None

                # print "PRE-INTERRUPT:",len_pre,d_slice_pre,m_slice_pre
                yield len_pre, d_slice_pre, m_slice_pre
                fpos += len_pre
                disk_interrupt_fn(disk_interrupt_points[i])

                # prepare for next interrupt

                i += 1
                if len(disk_interrupt_points) > i:
                    next_dip = disk_interrupt_points[i]
                else:
                    next_dip = None

                # update 'input' to reflect post reading slice (may still need further
                # decomposition)
                nread_disk, disk_slice, mem_slice = len_post, d_slice_post, m_slice_post

                # print "POST-INTERRUPT:",nread_disk, disk_slice,
                # mem_slice,"continues@",fpos

            yield nread_disk, disk_slice, mem_slice
            fpos += nread_disk

    def iterate(self, families_on_disk, families_in_memory, multiskip=False):
        """Provide an iterator which yields step-by-step instructions
        for partial-loading an array with the specified families stored
        on disk into a memory array containing the specified families.

        Each output consists of *readlen*, *file_index*, *memory_index*.
        A typical read loop should be as follows:

        for readlen, buffer_index, memory_index in ctl.iterate(fams_on_disk, fams_in_mem) :
          data = read_entries(count=readlen)
          if memory_index is not None :
            target_array[memory_index] = data[buffer_index]

        Obviously this can be optimized, for instance to skip through
        file data when memory_index is None rather than read and discard it.

        **Input:** :

          *families_on_disk*: list of families for which the array exists on disk
          *families_in_memory*: list of families to target in memory
          *multiskip*: if True, skip commands (i.e. entries with file_index=None)
             can have readlen greater than the block length

          """

        mem_offset = 0

        skip_accumulation = 0

        for current_family in self._ordered_families:
            if skip_accumulation > 0:
                yield skip_accumulation, None, None
                skip_accumulation = 0

            if current_family not in families_on_disk:
                assert current_family not in families_in_memory
            else:
                if current_family in families_in_memory:
                    for nread_disk, disk_mask, mem_slice in self._family_chunks[current_family]:
                        if mem_slice is None:
                            if multiskip:
                                skip_accumulation += nread_disk
                            else:
                                yield nread_disk, None, None
                        else:
                            if skip_accumulation > 0:
                                yield skip_accumulation, None, None
                                skip_accumulation = 0
                            mem_slice_offset = slice(
                                mem_slice.start + mem_offset, mem_slice.stop + mem_offset)
                            yield nread_disk, disk_mask, mem_slice_offset

                    mem_fs = self.mem_family_slice[current_family]
                    mem_offset += mem_fs.stop - mem_fs.start
                else:
                    for nread_disk, disk_mask, mem_slice in self._family_chunks[current_family]:
                        yield nread_disk, None, None



"""
interpolate
===========

2D and 3D Interpolation routines written in cython

"""

import numpy as np

from . import _interpolate3d

# this just calls the cython interpolation function, setting the
# interpolation arrays to correct type


def interpolate3d(x, y, z, x_vals, y_vals, z_vals, vals):
    """
    Interpolate on a 3D regular grid.
    Yields results identical to scipy.interpolate.interpn.

    Input
    -----

    x,y,z : points where the interpolation will be performed

    x_vals, y_vals, z_vals : xyz values of the reference grid

    vals : grid values
    """

    # cast x_vals, y_vals and z_vals to float64

    x_vals = x_vals.astype(np.float64)
    y_vals = y_vals.astype(np.float64)
    z_vals = z_vals.astype(np.float64)
    vals = vals.astype(np.float64)

    result_array = np.empty(len(x), dtype=np.float64)

    _interpolate3d.interpolate3d(len(x),
                                 x, y, z,
                                 len(x_vals), x_vals,
                                 len(y_vals), y_vals,
                                 len(z_vals), z_vals,
                                 vals,
                                 result_array)

    return result_array


def interpolate2d(x, y, x_vals, y_vals, vals):
    """
    Interpolate on a 2D regular grid.
    Yields results identical to scipy.interpolate.interpn.

    Input
    -----

    x,y : points where the interpolation will be performed

    x_vals, y_vals : xy values of the reference grid

    vals : grid values
    """

    x_vals = x_vals.astype(np.float64)
    y_vals = y_vals.astype(np.float64)
    z_vals = np.ndarray(1, dtype=np.float64)

    vals = vals.astype(np.float64)
    vals.resize((1,) + vals.shape)

    result_array = np.empty(len(x), dtype=np.float64)

    _interpolate3d.interpolate3d(len(x),
                                 np.ndarray(1, dtype=np.float64), x, y,
                                 0, np.ndarray(1, dtype=np.float64),
                                 len(x_vals), x_vals,
                                 len(y_vals), y_vals,
                                 vals,
                                 result_array)

    return result_array



from os import environ

import matplotlib.pylab as plt

import pynbody

# load the snapshot and set to physical units
s = pynbody.load('testdata/g15784.lr.01024.gz')

# load the halos
h = s.halos()

# center on the largest halo and align the disk
pynbody.analysis.angmom.faceon(h[1])

# create a profile object for the stars
s.physical_units()
p = pynbody.analysis.profile.Profile(h[1],rmin=.01,rmax =250,type='log',ndim=3)
pg = pynbody.analysis.profile.Profile(h[1].g,rmin=.01,rmax =250,type='log',ndim=3)
ps = pynbody.analysis.profile.Profile(h[1].s,rmin=.01,rmax =250,type='log',ndim=3)
pd = pynbody.analysis.profile.Profile(h[1].d,rmin=.01,rmax =250,type='log',ndim=3)

# make the plot
plt.plot(p['rbins'],p['v_circ'],label='total')
plt.plot(pg['rbins'],pg['v_circ'],label='gas')
plt.plot(ps['rbins'],ps['v_circ'],label='stars')
plt.plot(pd['rbins'],pd['v_circ'],label='dm')

plt.xlabel('R [kpc]')
plt.ylabel(r'$v_c$ [km/s]')
plt.legend()



"""

units
=====

The pynbody units module consists of a set of classes for tracking units.

It relates closely to the :mod:`~pynbody.array` module, which defines
an extension to numpy arrays which carries unit information.

Making units
------------

Units are generated and used at various points through the pynbody
framework. Quite often the functions where users interact with units
simply accept strings.

You can also make units yourself in two ways. Either you can create a string, and
instantiate a Unit like this:

>>> units.Unit("Msol kpc**-3")
>>> units.Unit("2.1e12 m_p cm**-2/3")

Or you can do it within python, using the named Unit objects

>>> units.Msol * units.kpc**-3
>>> 2.1e12 * units.m_p * units.cm**(-2,3)

In the last example, either a tuple describing a fraction or a
Fraction instance (from the standard python module fractions) is
acceptable.


Getting conversion ratios
-------------------------

To convert one unit to another, use the ``ratio`` member function:

>>> units.Msol.ratio(units.kg)
1.99e30
>>> (units.Msol / units.kpc**3).ratio(units.m_p/units.cm**3)
4.04e-8

If the units cannot be converted, a UnitsException is raised:

>>> units.Msol.ratio(units.kpc)
UnitsException

Specifying numerical values
---------------------------

Sometimes it's necessary to specify a numerical value in the course
of a conversion. For instance, consider a comoving distance; this
can be specified in pynbody units as follows:

>>> comoving_kpc = units.kpc * units.a

where units.a represents the scalefactor. We can attempt to convert
this to a physical distance as follows

>>> comoving_kpc.ratio(units.kpc)

but this fails, throwing a UnitsException. On the other hand we
can specify a value for the scalefactor when we request the conversion

>>> comoving_kpc.ratio(units.kpc, a=0.5)
0.5

and the conversion completes with the expected result. The units
module also defines units.h for the dimensionless hubble constant,
which can be used similarly. *By default, all conversions happening
within a specific simulation context should pass in values for
a and h as a matter of routine.*

Any IrreducibleUnit (see below) can have a value specified in this way,
but a and h are envisaged to be the most useful applications.

Defining new base units
-----------------------

The units module is fully extensible: you can define and name your own
units which then integrate with all the standard functions.

.. code-block:: python

   litre = units.NamedUnit("litre",0.001*units.m**3)
   gallon = units.NamedUnit("gallon",0.004546*units.m**3)
   gallon.ratio(litre) # 4.546
   (units.pc**3).ratio(litre) # 2.94e52


You can even define completely new dimensions.

.. code-block:: python

    V = units.IrreducibleUnit("V") # define a volt
    C = units.NamedUnit("C", units.J/V) # define a coulomb
    q = units.NamedUnit("q", 1.60217646e-19*C) # elementary charge
    F = units.NamedUnit("F", C/V) # Farad
    epsilon0 = 8.85418e-12 *F/units.m


>>> (q*V).ratio("eV")
1.000
>>> ((q**2)/(4*math.pi*epsilon0*units.m**2)).ratio("N")
2.31e-28


"""

import fractions
import functools
import keyword
import numbers
import re
from collections import defaultdict

import numpy as np

Fraction = fractions.Fraction

_registry = {}


class UnitsException(Exception):
    pass


class UnitBase:

    """Base class for units. To instantiate a unit, call the :func:`pynbody.units.Unit`
    factory function."""

    def __init__(self):
        raise ValueError("Cannot directly initialize abstract base class")
        pass

    def __pow__(self, p):
        if isinstance(p, tuple):
            p = Fraction(p[0], p[1])
        if not (isinstance(p, Fraction) or isinstance(p, int)):
            if isinstance(p, float):
                raise ValueError("Units can only be raised to integer or fractional powers. Use python's built-in fractions module or a tuple: e.g. unit**(1,2) represents a square root.")
            else :
                raise ValueError("Units can only be raised to integer or fractional powers")
        return CompositeUnit(1, [self], [p]).simplify()

    def __truediv__(self, m):
        return self.__div__(m)

    def __rtruediv__(self, m):
        return self.__rdiv__(m)

    def __div__(self, m):
        if hasattr(m, "_no_unit"):
            return NoUnit()

        if isinstance(m, UnitBase):
            return CompositeUnit(1, [self, m], [1, -1]).simplify()
        else:
            return CompositeUnit(1.0 / m, [self], [1]).simplify()

    def __rdiv__(self, m):
        return CompositeUnit(m, [self], [-1]).simplify()

    def __mul__(self, m):
        if hasattr(m, "_no_unit"):
            return NoUnit()
        elif hasattr(m, "units"):
            return m * self
        elif isinstance(m, UnitBase):
            return CompositeUnit(1, [self, m], [1, 1]).simplify()
        else:
            return CompositeUnit(m, [self], [1]).simplify()

    def __rmul__(self, m):
        return CompositeUnit(m, [self], [1]).simplify()

    def __add__(self, m):
        scale = m.in_units(self) if hasattr(m, 'in_units') else m/float(self)
        if hasattr(scale, 'units'):
            scale.units = 1
        return self * (1.0 + scale)

    def __sub__(self, m):
        return self + (-m)

    def __repr__(self):
        return 'Unit("' + str(self) + '")'

    def __eq__(self, other):
        if not isinstance(other, UnitBase):
            other = Unit(other)
        _self_scale = getattr(self, '_scale', None)
        _other_scale = getattr(other, '_scale', None)
        if self.dimensionality_as_string() != other.dimensionality_as_string():
            return False
        elif _self_scale == 0.0 and _other_scale == 0.0:
            return True
        elif _self_scale == 0.0 or _other_scale == 0.0:
            return False
        else:
            try:
                return self.ratio(other) == 1.
            except UnitsException:
                return False


    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.ratio(other) < 1.

    def __gt__(self, other):
        return self.ratio(other) > 1.

    def __le__(self, other):
        return self.ratio(other) <= 1.

    def __ge__(self, other):
        return self.ratio(other) >= 1.

    def __neg__(self):
        return self * (-1)

    def __float__(self):
        return 1.

    def __hash__(self):
        return id(self)

    def _dimension_state(self):
        state = defaultdict(int)
        for base, power in zip(self._bases, self._powers):
            if isinstance(base, UnitBase):
                for k, v in base._dimension_state().items():
                    state[k] += v * power
            else:
                state[base] += power

        return state

    def dimensionality_as_string(self):
        """
        Returns the dimensionality of the Unit object.

        Example
        -------
        > pynbody.units.Unit("3e8 m s**-1 yr").dimensionality_as_string()
        'm^1'
        """
        state = self._dimension_state()
        return " ".join(
            f"{k}^{state[k]}" for k in sorted(state) if state[k] != 0
        )

    def simplify(self):
        return self

    def is_dimensionless(self):
        return False

    def ratio(self, other, **substitutions):
        """Get the conversion ratio between this Unit and another
        specified unit.

        Keyword arguments, if specified, give numerical substitutions
        for the named unit. This is most useful for specifying values
        for cosmological quantities like 'a' and 'h', but can also
        be used for any IrreducibleUnit.

        >>> Unit("1 Mpc a").ratio("kpc", a=0.25)
        250.0
        >>> Unit("1 Mpc").ratio("Msol")
        UnitsException: not convertible
        >>> Unit("1 Mpc").ratio("Msol", kg=25.0, m=50.0)
        3.1028701506345152e-08
        """

        if isinstance(other, str):
            other = Unit(other)

        if hasattr(other, "_no_unit"):
            raise UnitsException("Unknown units")

        try:
            return (self / other).dimensionless_constant(**substitutions)
        except UnitsException:
            raise UnitsException("Not convertible")

    def in_units(self, *a, **kw):
        """Alias for ratio"""

        return self.ratio(*a, **kw)

    def irrep(self):
        """Return a unit equivalent to this one (may be identical) but
        expressed in terms of the currently defined IrreducibleUnit
        instances."""
        return self

    def _register_unit(self, st):
        if st in _registry:
            raise UnitsException("Unit with this name already exists")
        if "**" in st or "^" in st or " " in st:
            # will cause problems for simple string parser in Unit() factory
            raise UnitsException("Unit names cannot contain '**' or '^' or spaces")
        _registry[st] = self

    def __deepcopy__(self, memo):
        # This may look odd, but the units conversion will be very
        # broken after deep-copying if we don't guarantee that a given
        # physical unit corresponds to only one instance
        return self


class NoUnit(UnitBase):

    def __init__(self):
        self._no_unit = True

    def _dimension_state(self):
        return {}

    def ratio(self, other, **substitutions):
        if isinstance(other, NoUnit):
            return 1
        else:
            raise UnitsException("Unknown units")

    def dimensional_project(self, *args):
        raise UnitsException("Unknown units")

    def is_dimensionless(self):
        return True

    def simplify(self):
        return self

    def __pow__(self, a):
        return self

    def __div__(self, a):
        return self

    def __rdiv__(self, a):
        return self

    def __mul__(self, a):
        return self

    def __rmul__(self, a):
        return self

    def __repr__(self):
        return "NoUnit()"

    def latex(self):
        return ""

    def irrep(self):
        return self

no_unit = NoUnit()


def _resurrect_named_unit(unit_name, unit_latex, represents):
    if unit_name in _registry:
        return _registry[unit_name]
    else:
        if represents is None:
            nu = IrreducibleUnit(unit_name)
        else:
            nu = NamedUnit(unit_name, represents)
            nu._latex = unit_latex
        return nu


class IrreducibleUnit(UnitBase):

    def __init__(self, st):
        self._st_rep = st
        self._register_unit(st)
        self._bases = [st]
        self._powers = [1]

    def __reduce__(self):
        return (_resurrect_named_unit, (self._st_rep, None, None))

    def __str__(self):
        return self._st_rep

    def latex(self):
        return r"\mathrm{" + self._st_rep + "}"

    def irrep(self):
        return CompositeUnit(1, [self], [1])


class NamedUnit(UnitBase):

    def __init__(self, st, represents):
        self._st_rep = st
        if isinstance(represents, str):
            represents = Unit(represents)

        self._represents = represents
        self._register_unit(st)

    def __reduce__(self):
        return (_resurrect_named_unit, (self._st_rep, getattr(self, '_latex', None), self._represents))

    def __str__(self):
        return self._st_rep

    def _dimension_state(self):
        return self._represents._dimension_state()

    def latex(self):
        if hasattr(self, '_latex'):
            return self._latex
        return r"\mathrm{" + self._st_rep + "}"

    def irrep(self):
        return self._represents.irrep()


class CompositeUnit(UnitBase):

    def __init__(self, scale, bases, powers):
        """Initialize a composite unit.

        Direct use of this function is not recommended. Instead use the
        factory function Unit(...)."""

        if scale == 1.:
            scale = 1

        self._scale = scale
        self._bases = bases
        self._powers = powers

    def latex(self):
        r"""Returns a LaTeX representation of this unit.

        Prefactors are converted into exponent notation. Named units by default
        are represented by the string '\mathrm{unit_name}', although this can
        be overriden in the pynbody configuration files or by setting
        unit_name._latex."""

        if self._scale != 1:
            x = ("%.2e" % self._scale).split('e')
            s = x[0]
            ex = x[1].lstrip('0+')
            if len(ex) > 0 and ex[0] == '-':
                ex = '-' + (ex[1:]).lstrip('0')
            if ex != '':
                s += r"\times 10^{" + ex + "}"
        else:
            s = ""

        for b, p in zip(self._bases, self._powers):
            if s != "":
                s += r"\," + b.latex()
            else:
                s = b.latex()

            if p != 1:
                s += "^{"
                s += str(p)
                s += "}"
        return s

    def __str__(self):
        s = None
        if len(self._bases) == 0:
            return "%.2e" % self._scale

        if self._scale != 1:
            s = "%.2e" % self._scale

        for b, p in zip(self._bases, self._powers):
            if s is not None:
                s += " " + str(b)
            else:
                s = str(b)

            if p != 1:
                s += "**"
                if isinstance(p, Fraction):
                    s += str(p)
                else:
                    s += str(p)
        return s

    def __float__(self):
        return float(self._scale)

    def _expand(self, expand_to_irrep=False):
        """Internal routine to expand any pointers to composite units
        into direct pointers to the base units. If expand_to_irrep is
        True, everything is expressed in irreducible units.
        A _gather will normally be necessary to sanitize the unit
        after an _expand."""

        trash = []

        for i, (b, p) in enumerate(zip(self._bases, self._powers)):
            if isinstance(b, NamedUnit) and expand_to_irrep:
                b = b._represents.irrep()

            if isinstance(b, CompositeUnit):
                if expand_to_irrep:
                    b = b.irrep()

                trash.append(i)
                self._scale *= b._scale ** p
                for b_sub, p_sub in zip(b._bases, b._powers):
                    self._bases.append(b_sub)
                    self._powers.append(p_sub * p)

        trash.sort()
        for offset, i in enumerate(trash):
            del self._bases[i - offset]
            del self._powers[i - offset]

    def _gather(self):
        """Internal routine to gather together powers of the same base
        units, then order the base units by their power (descending)"""

        trash = []
        bases = list(set(self._bases))
        powers = [sum(p for bi, p in zip(self._bases, self._powers)
                       if bi is b)
                  for b in bases]

        bp = sorted((x for x in zip(powers, bases) if x[0] != 0),
                    reverse=True,
                    key=lambda x: x[0])

        if len(bp) != 0:
            self._powers, self._bases = list(map(list, list(zip(*bp))))
        else:
            self._powers, self._bases = [], []

    def copy(self):
        """Create a copy which is 'shallow' in the sense that it
        references exactly the same underlying base units, but where
        the list of those units can be manipulated separately."""
        return CompositeUnit(self._scale, self._bases[:], self._powers[:])

    def __copy__(self):
        """For compatibility with python copy module"""
        return self.copy()

    def simplify(self):
        self._expand()
        self._gather()
        return self

    def irrep(self):
        """Return a new unit which represents this unit expressed
        solely in terms of IrreducibleUnit bases."""
        x = self.copy()
        x._expand(True)
        x._gather()
        return x

    def is_dimensionless(self):
        """Returns true if this unit actually translates into a scalar
        quantity."""
        x = self.irrep()
        if len(x._powers) == 0:
            return True

    def dimensionless_constant(self, **substitutions):
        """If this unit is dimensionless, return its scalar quantity.

        Direct use of this function is not recommended. It is generally
        better to use the ratio function instead.

        Provide keyword arguments to set values for named IrreducibleUnits --
        see the ratio function for more information."""

        x = self.irrep()
        c = x._scale
        for xb, xp in zip(x._bases, x._powers):
            if str(xb) in substitutions:
                c *= substitutions[str(xb)] ** xp
            else:
                raise UnitsException("Not dimensionless")

        return c

    def _power_of(self, base):
        if base in self._bases:
            return self._powers[self._bases.index(base)]
        else:
            return 0

    def dimensional_project(self, basis_units):
        """Work out how to express the dimensions of this unit relative to the
        specified list of basis units.

        This is used by the framework when making inferences about sensible units to
        use in various situations.

        For example, you can represent a length as an energy divided by a force:

           >>> Unit("23 kpc").dimensional_project(["J", "N"])
           array([1, -1], dtype=object)

        However it's not possible to represent a length by energy alone:

           >>> Unit("23 kpc").dimensional_project(["J"])
           UnitsException: Basis units do not span dimensions of specified unit

        This function also doesn't know what to do if the result is ambiguous:

           >>> Unit("23 kpc").dimensional_project(["J", "N", "kpc"])
           UnitsException: Basis units are not linearly independent

        """

        vec_irrep = [Unit(x).irrep() for x in basis_units]
        me_irrep = self.irrep()
        bases = set(me_irrep._bases)
        for vec in vec_irrep:
            bases.update(vec._bases)

        bases = list(bases)

        matrix = np.zeros((len(bases), len(vec_irrep)), dtype=Fraction)

        for base_i, base in enumerate(bases):
            for vec_i, vec in enumerate(vec_irrep):
                matrix[base_i, vec_i] = vec._power_of(base)

        # The matrix calculated above describes the transformation M
        # such that v = M.d where d is the sought-after powers of the
        # specified base vectors, and v is the powers in terms of the
        # base units in the list bases.
        #
        # To invert, since M is possibly rectangular, we use the
        # solution to the least-squares problem [minimize (v-M.d)^2]
        # which is d = (M^T M)^(-1) M^T v.
        #
        # If the solution to that does not solve v = M.d, there is no
        # admissable solution to v=M.d, i.e. the supplied base vectors do not
        # span
        # the requires space.
        #
        # If (M^T M) is singular, the vectors are not linearly independent, so
        # any
        # solution would not be unique.
        M_T_M = np.dot(matrix.transpose(), matrix)

        from . import util

        try:
            M_T_M_inv = util.rational_matrix_inv(M_T_M)
        except np.linalg.linalg.LinAlgError:
            raise UnitsException("Basis units are not linearly independent")

        my_powers = [me_irrep._power_of(base) for base in bases]

        candidate = np.dot(M_T_M_inv, np.dot(matrix.transpose(), my_powers))

        # Because our method involves a loss of information (multiplying
        # by M^T), we could get a spurious solution. Check this is not the
        # case...

        if any(np.dot(matrix, candidate) != my_powers):
            # Spurious solution, meaning the base vectors did not span the
            # units required in the first place.
            raise UnitsException(
                "Basis units do not span dimensions of specified unit")

        return candidate


def Unit(s):
    """
    Class factory for units. Given a string s, creates
    a Unit object.

    The string format is:
      [<scale>] [<unit_name>][**<rational_power>] [[<unit_name>] ... ]

    for example:

      "1.e30 kg"

      "kpc**2"

      "26.2 m s**-1"
    """

    if isinstance(s, UnitBase):
        return s
    elif isinstance(s, numbers.Number):
        scale = float(s)
        units = []
        powers = []
    else:
        s = str(s)

        x = s.split()
        try:
            scale = float(x[0])
            del x[0]
        except (ValueError, IndexError):
            scale = 1.0

        units = []
        powers = []

        for com in x:
            if "**" in com or "^" in com:
                s = com.split("**" if "**" in com else "^")
                try:
                    u = _registry[s[0]]
                except KeyError:
                    raise ValueError("Unknown unit " + s[0])
                p = Fraction(s[1])
                if p.denominator == 1:
                    p = p.numerator
            else:
                u = _registry[com]
                p = 1

            units.append(u)
            powers.append(p)

    return CompositeUnit(scale, units, powers)


def takes_arg_in_units(*args, **orig_kwargs):
    """

    Returns a decorator to create a function which auto-converts input
    to given units.

    **Usage:**

    .. code-block:: python

        @takes_arg_in_units((2, "Msol"), (1, "kpc"), ("blob", "erg"))
        def my_function(arg0, arg1, arg2, blob=22) :
           print "Arg 2 is",arg2,"Msol"
           print "Arg 1 is",arg1,"kpc"
           print "blob is",blob,"ergs"



    >>> My_function(22, "1.e30 kg", 23, blob="12 J")
    Input 3 is 0.5 Msol
    Input 2 is 23 kpc

    """

    context_arg = orig_kwargs.get('context_arg', None)

    kwargs = [x for x in args if hasattr(x[0], '__len__')]
    args = [x for x in args if not hasattr(x[0], '__len__')]

    def decorator_fn(x):
        @functools.wraps(x)
        def wrapper_fn(*fn_args, **fn_kwargs):
            context = {}
            if context_arg is not None:
                context = fn_args[context_arg].conversion_context()

            fn_args = list(fn_args)

            for arg_num, arg_units in args:

                if isinstance(fn_args[arg_num], str):
                    fn_args[arg_num] = Unit(fn_args[arg_num])

                if hasattr(fn_args[arg_num], "in_units"):
                    fn_args[arg_num] = fn_args[
                        arg_num].in_units(arg_units, **context)

            for arg_name, arg_units in kwargs:
                if isinstance(fn_kwargs[arg_name], str):
                    fn_kwargs[arg_name] = Unit(fn_kwargs[arg_name])

                if hasattr(fn_kwargs[arg_name], "in_units"):
                    fn_kwargs[arg_name] = fn_kwargs[
                        arg_name].in_units(arg_units, **context)

            return x(*fn_args, **fn_kwargs)

        return wrapper_fn

    return decorator_fn


from . import config_parser


def __is_clean_name(s):
    if re.search('[^0-9a-zA-Z_]', s):
        return False
    if re.search('^[^a-zA-Z_]+', s):
        return False
    if keyword.iskeyword(s):
        return False
    return True

for new_unit_name in map(str.strip, config_parser.get('irreducible-units', 'names').split(",")):
    new_unit = IrreducibleUnit(new_unit_name)
    if __is_clean_name(new_unit_name):
        globals()[new_unit_name] = new_unit

for new_unit_name, new_unit_definition in config_parser.items("named-units"):
    new_unit = NamedUnit(new_unit_name, new_unit_definition)
    if __is_clean_name(new_unit_name):
        globals()[new_unit_name] = new_unit

for unit_name, latex in config_parser.items("units-latex"):
    _registry[unit_name]._latex = latex


_default_units = {}

for a_, b_ in config_parser.items("default-array-dimensions"):
    _default_units[a_] = Unit(b_)


def has_unit(obj):
    """Returns True if the specified object has a meaningful units attribute"""
    if hasattr(obj, 'units') and isinstance(obj.units, UnitBase):
        return not hasattr(obj.units, '_no_unit')
    else:
        return False

has_units = has_unit

def get_item_with_unit(array, item):
    if has_unit(array) and len(array.shape)==1:
        return array[item]*array.units
    else:
        return array[item]


def is_unit(obj):
    """Returns True if the specified object represents a unit"""

    return isinstance(obj, UnitBase)


def is_unit_like(obj):
    """Returns True if the specified object is itself a unit or
    otherwise exposes unit information"""

    return is_unit(obj) or has_unit(obj)



"""

family
======

This module defines the Family class which represents
families of particles (e.g. dm, gas, star particles).
New Family objects are automatically registered so that
snapshots can use them in the normal syntax (snap.dm,
snap.star, etc).

In practice the easiest way to make use of the flexibility
this module provides is through adding more families of
particles in your config.ini.

"""

import functools
import sys

from . import config_parser

_registry = []


def family_names(with_aliases=False):
    """Returns a list of the names of all particle families.
    If with_aliases is True, include aliases in the list."""

    global _registry
    l = []
    for o in _registry:
        l.append(o.name)
        if with_aliases:
            for a in o.aliases:
                l.append(a)
    return l


def get_family(name, create=False):
    """Returns a family corresponding to the specified string.  If the
    family does not exist and create is False, raises ValueError. If
    the family does not exist and create is True, an appropriate
    object is instantiated, registered and returned."""

    if isinstance(name, Family):
        return name

    name = name.lower()
    # or should it check and raise rather than just convert?
    # Not sure.
    for n in _registry:
        if n.name == name or name in n.aliases:
            return n

    if create:
        return Family(name)
    else:
        raise ValueError(name +
                         " not a family")  # is ValueError the right thing here?


class Family:

    def __init__(self, name, aliases=[]):
        if name != name.lower():
            raise ValueError("Family names must be lower case")
        if name in family_names(with_aliases=True):
            raise ValueError("Family name " + name + " is not unique")
        for a in aliases:
            if a != a.lower():
                raise ValueError("Aliases must be lower case")

        self.name = name
        self.aliases = aliases
        _registry.append(self)

    def __repr__(self):
        return "<Family " + self.name + ">"

    def __reduce__(self):
        return get_family, (self.name, True), {"aliases": self.aliases}

    def __iter__(self):
        # Provided so a single family can be treated as a list of families
        yield self

    def __str__(self):
        return self.name

    def __cmp__(self, other):
        # for python 2.x
        return cmp(str(self), str(other))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(str(self))

if sys.version_info[0] >= 3:
    Family = functools.total_ordering(Family)


# Instantiate the default families as specified
# by the configuration file

g = globals()
for f in config_parser.options('families'):
    aliases = config_parser.get('families', f)
    g[f] = Family(f, list(map(str.strip, aliases.split(","))))



"""

pkdgrav_cosmo
=============

Cosmological module from PKDGRAV.

N.B.  This code is being shared with skid and the I.C. generator.

**NEEDS DOCUMENTATION**
"""

import math

from scipy.integrate import ode, romberg


class Cosmology:

    """ docs placeholder """

    EPSCOSMO = 1e-7

    def __init__(self, sim=None, H0=math.sqrt(8 * math.pi / 3), Om=0.272,
                 L=0.728, Ob=0.0456,
                 Or=0.0, Quin=0.0, Ok=0.0):
        if sim is not None:
            self.dOmegaM = sim.properties['omegaM0']
            self.dLambda = sim.properties['omegaL0']
        else:
            self.dOmegaM = Om
            self.dLambda = L
        self.dHubble0 = H0
        self.dOmegab = Ob
        self.dOmegaRad = Or
        self.dQuintess = Quin
        self.dOmegaCurve = Ok
        self.bComove = 1

    # The cosmological equation of state is entirely determined here.  We
    # will derive all other quantities from these two functions.
    def Exp2Hub(self, dExp):
        dOmegaCurve = (1.0 - self.dOmegaM - self.dLambda -
                       self.dOmegaRad - self.dQuintess)

        assert(dExp > 0.0)
        return (self.dHubble0 * math.sqrt(self.dOmegaM * dExp
                                          + dOmegaCurve * dExp * dExp
                                          + self.dOmegaRad
                                          + self.dQuintess * dExp *
                                          dExp * math.sqrt(dExp)
                                          + self.dLambda * dExp * dExp * dExp * dExp) / (dExp * dExp))

    # Return a double dot over a.
    def ExpDot2(self, dExp):
        return (self.dHubble0 * self.dHubble0 *
                (self.dLambda - 0.5 * self.dOmegaM / (dExp * dExp * dExp)
                 + 0.25 * self.dQuintess / (dExp * math.sqrt(dExp))
                 - self.dOmegaRad / (dExp * dExp * dExp * dExp)))

    def Time2Hub(self, dTime):
        a = self.Time2Exp(dTime)
        assert(a > 0.0)
        return self.Exp2Hub(a)

    def CosmoTint(self, dY):
        dExp = dY ** (2.0 / 3.0)
        assert(dExp > 0.0)
        return 2.0 / (3.0 * dY * self.Exp2Hub(dExp))

    def Exp2Time(self, dExp):
        dOmegaM = self.dOmegaM
        dHubble0 = self.dHubble0

        if(self.dLambda == 0.0 and self.dOmegaRad == 0.0 and
           self.dQuintess == 0.0):
            if (dOmegaM == 1.0):
                assert(dHubble0 > 0.0)
                if (dExp == 0.0):
                    return(0.0)
                return(2.0 / (3.0 * dHubble0) * dExp ** 1.5)

            elif (dOmegaM > 1.0):
                assert(dHubble0 >= 0.0)
                if (dHubble0 == 0.0):
                    B = 1.0 / math.sqrt(dOmegaM)
                    eta = acos(1.0 - dExp)
                    return(B * (eta - sin(eta)))

                if (dExp == 0.0):
                    return(0.0)

                a0 = 1.0 / dHubble0 / math.sqrt(dOmegaM - 1.0)
                A = 0.5 * dOmegaM / (dOmegaM - 1.0)
                B = A * a0
                eta = acos(1.0 - dExp / A)
                return(B * (eta - sin(eta)))

            elif (dOmegaM > 0.0):
                assert(dHubble0 > 0.0)
                if (dExp == 0.0):
                    return(0.0)
                a0 = 1.0 / dHubble0 / math.sqrt(1.0 - dOmegaM)
                A = 0.5 * dOmegaM / (1.0 - dOmegaM)
                B = A * a0
                eta = acosh(dExp / A + 1.0)
                return(B * (sinh(eta) - eta))

            elif (dOmegaM == 0.0):
                assert(dHubble0 > 0.0)
                if (dExp == 0.0):
                    return(0.0)
                return(dExp / dHubble0)

            else:
                #* Bad value.
                assert(0)
                return(0.0)

        else:
            # Set accuracy to 0.01 EPSCOSMO to make Romberg integration
            # more accurate than Newton's method criterion in Time2Exp. --JPG
            return romberg(self.CosmoTint, self.EPSCOSMO, dExp ** 1.5, tol=0.01 * self.EPSCOSMO)

    def Time2Exp(self, dTime):
        dHubble0 = self.dHubble0

        dExpOld = 0.0
        dExpNew = dTime * dHubble0
        dDeltaOld = dExpNew
        # old change in interval
        dUpper = 1.0e38
        # bounds on root
        dLower = 0.0

        it = 0

        # Root find with Newton's method.
        while (math.fabs(dExpNew - dExpOld) / dExpNew > self.EPSCOSMO):
            f = dTime - self.Exp2Time(dExpNew)
            fprime = 1.0 / (dExpNew * self.Exp2Hub(dExpNew))

            if(f * fprime > 0):
                dLower = dExpNew
            else:
                dUpper = dExpNew

            dExpOld = dExpNew
            dDeltaOld = f / fprime
            dExpNext = dExpNew + dDeltaOld
            # check if bracketed
            if((dExpNext > dLower) and (dExpNext < dUpper)):
                dExpNew = dExpNext
            else:
                dExpNew = 0.5 * (dUpper + dLower)
            it += 1
            assert(it < 40)

        return dExpNew

    def ComoveDriftInt(self, dIExp):
        return -dIExp / (Exp2Hub(1.0 / dIExp))

#* Make the substitution y = 1/a to integrate da/(a^2*H(a))
    def ComoveKickInt(self, dIExp):
        return -1.0 / (self.Exp2Hub(1.0 / dIExp))

#* This function integrates the time dependence of the "drift"-Hamiltonian.
    def ComoveDriftFac(self, dTime, dDelta):
        dOmegaM = self.dOmegaM
        dHubble0 = self.dHubble0

        if(self.dLambda == 0.0 and self.dOmegaRad == 0.0 and
           self.dQuintess == 0.0):
            a1 = self.Time2Exp(dTime)
            a2 = self.Time2Exp(dTime + dDelta)
            if (dOmegaM == 1.0):
                return((2.0 / dHubble0) * (1.0 / math.sqrt(a1) - 1.0 / math.sqrt(a2)))

            elif (dOmegaM > 1.0):
                assert(dHubble0 >= 0.0)
                if (dHubble0 == 0.0):
                    A = 1.0
                    B = 1.0 / math.sqrt(dOmegaM)

                else:
                    a0 = 1.0 / dHubble0 / math.sqrt(dOmegaM - 1.0)
                    A = 0.5 * dOmegaM / (dOmegaM - 1.0)
                    B = A * a0

                    eta1 = acos(1.0 - a1 / A)
                    eta2 = acos(1.0 - a2 / A)
                    return(B / A / A * (1.0 / tan(0.5 * eta1) - 1.0 / tan(0.5 * eta2)))

            elif (dOmegaM > 0.0):
                assert(dHubble0 > 0.0)
                a0 = 1.0 / dHubble0 / math.sqrt(1.0 - dOmegaM)
                A = 0.5 * dOmegaM / (1.0 - dOmegaM)
                B = A * a0
                eta1 = acosh(a1 / A + 1.0)
                eta2 = acosh(a2 / A + 1.0)
                return(B / A / A * (1.0 / tanh(0.5 * eta1) - 1.0 / tanh(0.5 * eta2)))

            elif (dOmegaM == 0.0):
                # YOU figure this one out!
                assert(0)
                return(0.0)

            else:
                # Bad value?
                assert(0)
                return(0.0)
        else:
            return romberg(self.ComoveDriftInt, 1.0 / self.Time2Exp(dTime),
                           1.0 / self.Time2Exp(dTime + dDelta), tol=self.EPSCOSMO)

    # This function integrates the time dependence of the "kick"-Hamiltonian.
    def ComoveKickFac(self, dTime, dDelta):
        dOmegaM = self.dOmegaM
        dHubble0 = self.dHubble0

        if (not self.bComove):
            return(dDelta)
        elif(self.dLambda == 0.0 and self.dOmegaRad == 0.0
                and self.dQuintess == 0.0):
            a1 = self.Time2Exp(dTime)
            a2 = self.Time2Exp(dTime + dDelta)
            if (dOmegaM == 1.0):
                return((2.0 / dHubble0) * (math.sqrt(a2) - math.sqrt(a1)))

            elif (dOmegaM > 1.0):
                assert(dHubble0 >= 0.0)
                if (dHubble0 == 0.0):
                    A = 1.0
                    B = 1.0 / math.sqrt(dOmegaM)

                else:
                    a0 = 1.0 / dHubble0 / math.sqrt(dOmegaM - 1.0)
                    A = 0.5 * dOmegaM / (dOmegaM - 1.0)
                    B = A * a0

                eta1 = acos(1.0 - a1 / A)
                eta2 = acos(1.0 - a2 / A)
                return(B / A * (eta2 - eta1))

            elif (dOmegaM > 0.0):
                assert(dHubble0 > 0.0)
                a0 = 1.0 / dHubble0 / math.sqrt(1.0 - dOmegaM)
                A = 0.5 * dOmegaM / (1.0 - dOmegaM)
                B = A * a0
                eta1 = acosh(a1 / A + 1.0)
                eta2 = acosh(a2 / A + 1.0)
                return(B / A * (eta2 - eta1))

            elif (dOmegaM == 0.0):
                #* YOU figure this one out!
                assert(0)
                return(0.0)

            else:
                #* Bad value?
                assert(0)
                return(0.0)

        else:
            return romberg(self.ComoveKickInt, 1.0 / self.Time2Exp(dTime),
                           1.0 / self.Time2Exp(dTime + dDelta), tol=self.EPSCOSMO)

    def ComoveLookbackTime2Exp(self, dComoveTime):
        if (not self.bComove):
            return(1.0)
        else:
            dExpOld = 0.0
            dT0 = self.Exp2Time(1.0)
            dTime = dT0 - dComoveTime
            dExpNew
            it = 0

            if(dTime < self.EPSCOSMO):
                dTime = self.EPSCOSMO
            dExpNew = self.Time2Exp(dTime)
            # Root find with Newton's method.
            while (fabs(dExpNew - dExpOld) / dExpNew > self.EPSCOSMO):
                dTimeNew = self.Exp2Time(dExpNew)
                f = (dComoveTime - self.ComoveKickFac(
                    dTimeNew, dT0 - dTimeNew))
                fprime = -1.0 / (dExpNew * dExpNew * self.Exp2Hub(dExpNew))
                dExpOld = dExpNew
                dExpNew += f / fprime
                it += 1
                assert(it < 20)

        return dExpNew

# delta[1] => deltadot
    def GrowthFacDeriv(self, dlnExp, dlnDelta, dlnDeltadot):
        dExp = exp(dlnExp)
        dHubble = self.Exp2Hub(dExp)

        dlnDeltadot[0] = dlnDelta[1]
        dlnDeltadot[1] = (-dlnDelta[1] * dlnDelta[1]
                          - dlnDelta[1] * (1.0 + self.ExpDot2(
                              dExp) / (dHubble * dHubble))
                          + 1.5 * self.Exp2Om(dExp))

    def GrowthFac(self, dExp):
        dlnExpStart = -15
        nSteps = 200
        dlnExp = math.log(dExp)

        assert(dlnExp > dlnExpStart)

        dDStart[0] = dlnExpStart
        dDStart[1] = 1.0
        # Growing mode

        integrator = ode(GrowthFacDeriv).set_integrator('dopri5')
#        , 2,
#        , dDEnd,  nSteps);
        integrator = integrator.set_initial_value(
            2, dlnExpStart, dDStart, dlnExp)
        dDEnd = integrator.integrate(t1, step=0, relax=0)
        flag = integrator.successful()
        return exp(dDEnd[0])

    def GrowthFacDot(self, dExp):
        dlnExpStart = -15
        nSteps = 200
        dlnExp = math.log(dExp)
        dDStart[2]
        dDEnd[2]

        assert(dlnExp > dlnExpStart)

        dDStart[0] = dlnExpStart
        dDStart[1] = 1.0
        # Growing mode

        integrator = ode(GrowthFacDeriv).set_integrator('dopri5')
#        , 2, dlnExpStart, dDStart, dlnExp, dDEnd,nSteps);
        return dDEnd[1] * self.Exp2Hub(dExp) * exp(dDEnd[0])

# expansion dependence of Omega_matter
    def Exp2Om(self, dExp):
        dHubble = self.Exp2Hub(dExp)
        return (self.dOmegaM * self.dHubble0 * self.dHubble0 /
                (dExp * dExp * dExp * dHubble * dHubble))



"""
snapshot
========

This module implements the  :class:`~pynbody.snapshot.SimSnap` class which manages and stores snapshot data.
It also implements the :class:`~pynbody.snapshot.SubSnap` class (and relatives) which
represent different views of an existing :class:`~pynbody.snapshot.SimSnap`.

"""

import logging
import pathlib

from .. import config, family
from . import util
from .simsnap import SimSnap

logger = logging.getLogger('pynbody.snapshot')


def load(filename, *args, **kwargs) -> SimSnap:
    """Loads a file using the appropriate class, returning a SimSnap
    instance."""

    filename = pathlib.Path(filename)

    priority = kwargs.pop('priority', config['snap-class-priority'])

    for c in SimSnap.iter_subclasses_with_priority(priority):
        if c._can_load(filename):
            logger.info("Loading using backend %s" % str(c))
            return c(filename, *args, **kwargs)

    raise OSError(
        "File %r: format not understood or does not exist" % filename)

def new(n_particles=0, order=None, class_=SimSnap, **families):
    """Create a blank SimSnap, with the specified number of particles.

    Position, velocity and mass arrays are created and filled
    with zeros.

    By default all particles are taken to be dark matter.
    To specify otherwise, pass in keyword arguments specifying
    the number of particles for each family, e.g.

    f = new(dm=50, star=25, gas=25)

    The order in which the different families appear in the snapshot
    is unspecified unless you add an 'order' argument:

    f = new(dm=50, star=25, gas=25, order='star,gas,dm')

    guarantees the stars, then gas, then dark matter particles appear
    in sequence.
    """

    if len(families) == 0:
        families = {'dm': n_particles}

    t_fam = []
    tot_particles = 0

    if order is None:
        for k, v in list(families.items()):

            assert isinstance(v, int)
            t_fam.append((family.get_family(k), v))
            tot_particles += v
    else:
        for k in order.split(","):
            v = families[k]
            assert isinstance(v, int)
            t_fam.append((family.get_family(k), v))
            tot_particles += v

    x = class_()
    x._num_particles = tot_particles
    x._filename = "<created>"

    x._create_arrays(["pos", "vel"], 3)
    x._create_arrays(["mass"], 1)

    rt = 0
    for k, v in t_fam:
        x._family_slice[k] = slice(rt, rt + v)
        rt += v

    x._decorate()
    return x


from . import ascii, gadget, gadgethdf, grafic, nchilada, ramses, subsnap, tipsy
from .subsnap import FamilySubSnap, IndexedSubSnap, SubSnap



import numpy as np

from pynbody import filt, util
from pynbody.snapshot import SimSnap


class ExposedBaseSnapshotMixin:
    # The following will be objects common to a SimSnap and all its SubSnaps
    _inherited = ["_immediate_cache_lock",
                  "lazy_off", "lazy_derive_off", "lazy_load_off", "auto_propagate_off",
                  "properties", "_derived_array_names", "_family_derived_array_names",
                  "_dependency_tracker", "immediate_mode", "delay_promotion"]

    def __init__(self, base, *args, **kwargs):
        self.base = base
        super().__init__(base, *args, **kwargs)

    def _inherit(self):
        self._file_units_system = self.base._file_units_system
        self._unifamily = self.base._unifamily

        for x in self._inherited:
            setattr(self, x, getattr(self.base, x))

class SubSnapBase(SimSnap):
    def __init__(self, base):
        self._subsnap_base = base

    def _get_array(self, name, index=None, always_writable=False):
        if self.immediate_mode:
            return self._get_from_immediate_cache(name,
                                                  lambda: self._subsnap_base._get_array(
                                                      name, None, always_writable)[self._slice])

        else:
            ret = self._subsnap_base._get_array(name, util.concatenate_indexing(
                self._slice, index), always_writable)
            ret.family = self._unifamily
            return ret

    def _set_array(self, name, value, index=None):
        self._subsnap_base._set_array(
            name, value, util.concatenate_indexing(self._slice, index))

    def _get_family_array(self, name, fam, index=None, always_writable=False):
        base_family_slice = self._subsnap_base._get_family_slice(fam)
        sl = util.relative_slice(base_family_slice,
                                 util.intersect_slices(self._slice, base_family_slice, len(self._subsnap_base)))
        sl = util.concatenate_indexing(sl, index)
        if self.immediate_mode:
            return self._get_from_immediate_cache((name, fam),
                                                  lambda: self._subsnap_base._get_family_array(
                                                      name, fam, None, always_writable)[sl])
        else:
            return self._subsnap_base._get_family_array(name, fam, sl, always_writable)

    def _set_family_array(self, name, family, value, index=None):
        fslice = self._get_family_slice(family)
        self._subsnap_base._set_family_array(
            name, family, value, util.concatenate_indexing(fslice, index))

    def _promote_family_array(self, *args, **kwargs):
        self._subsnap_base._promote_family_array(*args, **kwargs)

    def __delitem__(self, name):
        # is this the right behaviour?
        raise RuntimeError("Arrays can only be deleted from the base snapshot")

    def _del_family_array(self, name, family):
        # is this the right behaviour?
        raise RuntimeError("Arrays can only be deleted from the base snapshot")

    @property
    def _filename(self):
        return str(self._subsnap_base._filename) + ":" + self._descriptor

    def keys(self):
        return list(self._subsnap_base.keys())

    def loadable_keys(self, fam=None):
        if self._unifamily:
            return self._subsnap_base.loadable_keys(self._unifamily)
        else:
            return self._subsnap_base.loadable_keys(fam)

    def derivable_keys(self):
        return self._subsnap_base.derivable_keys()

    def infer_original_units(self, *args):
        """Return the units on disk for a quantity with the specified dimensions"""
        return self._subsnap_base.infer_original_units(*args)

    def _get_family_slice(self, fam):
        sl = util.relative_slice(self._slice,
                                 util.intersect_slices(self._slice, self._subsnap_base._get_family_slice(fam), len(self._subsnap_base)))
        return sl

    def _load_array(self, array_name, fam=None, **kwargs):
        self._subsnap_base._load_array(array_name, fam)

    def write_array(self, array_name, fam=None, **kwargs):
        fam = fam or self._unifamily
        if not fam or self._get_family_slice(fam) != slice(0, len(self)):
            raise OSError(
                "Array writing is available for entire simulation arrays or family-level arrays, but not for arbitrary subarrays")

        self._subsnap_base.write_array(array_name, fam=fam, **kwargs)

    def _derive_array(self, array_name, fam=None):
        self._subsnap_base._derive_array(array_name, fam)

    def family_keys(self, fam=None):
        return self._subsnap_base.family_keys(fam)

    def _create_array(self, *args, **kwargs):
        self._subsnap_base._create_array(*args, **kwargs)

    def _create_family_array(self, *args, **kwargs):
        self._subsnap_base._create_family_array(*args, **kwargs)

    def physical_units(self, *args, **kwargs):
        self._subsnap_base.physical_units(*args, **kwargs)

    def is_derived_array(self, v, fam=None):
        fam = fam or self._unifamily
        return self._subsnap_base.is_derived_array(v, fam)

    def unlink_array(self, name):
        self._subsnap_base.unlink_array(name)

    def get_index_list(self, relative_to, of_particles=None):
        if of_particles is None:
            of_particles = np.arange(len(self))

        if relative_to is self:
            return of_particles

        return self._subsnap_base.get_index_list(relative_to, util.concatenate_indexing(self._slice, of_particles))

class SubSnap(ExposedBaseSnapshotMixin, SubSnapBase):
    """Represent a sub-view of a SimSnap, initialized by specifying a
    slice.  Arrays accessed through __getitem__ are automatically
    sub-viewed using the given slice."""

    def __init__(self, base, _slice):
        super().__init__(base)
        self._inherit()

        if isinstance(_slice, slice):
            # Various slice logic later (in particular taking
            # subsnaps-of-subsnaps) requires having positive
            # (i.e. start-relative) slices, so if we have been passed a
            # negative (end-relative) index, fix that now.
            if _slice.start is None:
                _slice = slice(0, _slice.stop, _slice.step)
            if _slice.start < 0:
                _slice = slice(len(
                    base) + _slice.start, _slice.stop, _slice.step)
            if _slice.stop is None or _slice.stop > len(base):
                _slice = slice(_slice.start, len(base), _slice.step)
            if _slice.stop < 0:
                _slice = slice(_slice.start, len(
                    base) + _slice.stop, _slice.step)

            self._slice = _slice

            descriptor = "[" + str(_slice.start) + ":" + str(_slice.stop)
            if _slice.step is not None:
                descriptor += ":" + str(_slice.step)
            descriptor += "]"

        else:
            raise TypeError("Unknown SubSnap slice type")

        self._num_particles = util.indexing_length(_slice)

        self._descriptor = descriptor






class IndexingViewMixin:
    def __init__(self, *args, **kwargs):
        index_array = kwargs.pop('index_array', None)
        iord_array = kwargs.pop('iord_array', None)
        super().__init__(*args, **kwargs)
        self._descriptor = "indexed"

        self._unifamily = self._subsnap_base._unifamily
        self._file_units_system = self._subsnap_base._file_units_system

        if index_array is None and iord_array is None:
            raise ValueError(
                "Cannot define a subsnap without an index_array or iord_array.")
        if index_array is not None and iord_array is not None:
            raise ValueError(
                "Cannot define a subsnap without both and index_array and iord_array.")
        if iord_array is not None:
            index_array = self._iord_to_index(iord_array)

        if isinstance(index_array, filt.Filter):
            self._descriptor = index_array._descriptor
            index_array = index_array.where(self._subsnap_base)[0]

        elif isinstance(index_array, tuple):
            if isinstance(index_array[0], np.ndarray):
                index_array = index_array[0]
            else:
                index_array = np.array(index_array)
        else:
            index_array = np.asarray(index_array)

        findex = self._subsnap_base._family_index()[index_array]
        # Check the family index array is monotonically increasing
        # If not, the family slices cannot be implemented
        if not all(np.diff(findex) >= 0):
            raise ValueError(
                "Families must retain the same ordering in the SubSnap")

        self._slice = index_array
        self._family_slice = {}
        self._family_indices = {}
        self._num_particles = len(index_array)

        # Find the locations of the family slices
        for i, fam in enumerate(self._subsnap_base.ancestor.families()):
            ids = np.where(findex == i)[0]
            if len(ids) > 0:
                new_slice = slice(ids.min(), ids.max() + 1)
                self._family_slice[fam] = new_slice
                self._family_indices[fam] = np.asarray(index_array[
                                                       new_slice]) - self._subsnap_base._get_family_slice(fam).start
    def _iord_to_index(self, iord):
        # Maps iord to indices. Note that this requires to perform an argsort (O(N log N) operations)
        # and a binary search (O(M log N) operations) with M = len(iord) and N = len(self._subsnap_base).

        if not util.is_sorted(iord) == 1:
            raise Exception('Expected iord to be sorted in increasing order.')

        # Find index of particles using a search sort
        iord_base = self._subsnap_base['iord']
        iord_base_argsort = self._subsnap_base['iord_argsort']
        index_array = util.binary_search(iord, iord_base, sorter=iord_base_argsort)

        # Check that the iord match
        if np.any(index_array == len(iord_base)):
            raise Exception('Some of the requested ids cannot be found in the dataset.')

        return index_array


class IndexedSubSnap(IndexingViewMixin, ExposedBaseSnapshotMixin, SubSnapBase):
    """Represents a subset of the simulation particles according
        to an index array.

        Parameters
        ----------
        base : SimSnap object
            The base snapshot
        index_array : integer array or None
            The indices of the elements that define the sub snapshot. Set to None to use iord-based instead.
        iord_array : integer array or None
            The iord of the elements that define the sub snapshot. Set to None to use index-based instead.
            This may be computationally expensive. See note below.

        Notes
        -----
        `index_array` and `iord_array` arguments are mutually exclusive.
        In the case of `iord_array`, an sorting operation is required that may take
        a significant time and require O(N) memory.
        """

    def __init__(self, base, index_array=None, iord_array=None, *args, **kwargs):
        super().__init__(base, index_array=index_array, iord_array=iord_array, *args, **kwargs)
        self._inherit()

    def _get_family_slice(self, fam):
        # A bit messy: jump out the SubSnap inheritance chain
        # and call SimSnap method directly...
        return SimSnap._get_family_slice(self, fam)

    def _get_family_array(self, name, fam, index=None, always_writable=False):
        sl = self._family_indices.get(fam,slice(0,0))
        sl = util.concatenate_indexing(sl, index)

        return self._subsnap_base._get_family_array(name, fam, sl, always_writable)

    def _set_family_array(self, name, family, value, index=None):
        self._subsnap_base._set_family_array(name, family, value,
                                    util.concatenate_indexing(self._family_indices[family], index))

    def _create_array(self, *args, **kwargs):
        self._subsnap_base._create_array(*args, **kwargs)


class FamilySubSnap(SubSnap):

    """Represents a one-family portion of a parent snap object"""

    def __init__(self, base, fam):
        super().__init__(base, base._get_family_slice(fam))

        self._unifamily = fam
        self._descriptor = ":" + fam.name


    def __delitem__(self, name):
        if name in list(self._subsnap_base.keys()):
            raise ValueError(
                "Cannot delete global simulation property from sub-view")
        elif name in self._subsnap_base.family_keys(self._unifamily):
            self._subsnap_base._del_family_array(name, self._unifamily)

    def keys(self):
        global_keys = list(self._subsnap_base.keys())
        family_keys = self._subsnap_base.family_keys(self._unifamily)
        return list(set(global_keys).union(family_keys))

    def family_keys(self, fam=None):
        # We now define there to be no family-specific subproperties,
        # because all properties can be accessed through standard
        # __setitem__, __getitem__ methods
        return []

    def _get_family_slice(self, fam):
        if fam is self._unifamily:
            return slice(0, len(self))
        else:
            return slice(0, 0)

    def _get_array(self, name, index=None, always_writable=False):
        try:
            return SubSnap._get_array(self, name, index, always_writable)
        except KeyError:
            return self._subsnap_base._get_family_array(name, self._unifamily, index, always_writable)

    def _create_array(self, array_name, ndim=1, dtype=None, zeros=True, derived=False, shared=None):
        # Array creation now maps into family-array creation in the parent
        self._subsnap_base._create_family_array(
            array_name, self._unifamily, ndim, dtype, derived, shared)

    def _set_array(self, name, value, index=None):
        if name in list(self._subsnap_base.keys()):
            self._subsnap_base._set_array(
                name, value, util.concatenate_indexing(self._slice, index))
        else:
            self._subsnap_base._set_family_array(name, self._unifamily, value, index)

    def _create_family_array(self, array_name, family, ndim, dtype, derived, shared):
        self._subsnap_base._create_family_array(
            array_name, family, ndim, dtype, derived, shared)

    def _promote_family_array(self, *args, **kwargs):
        pass

    def _load_array(self, array_name, fam=None, **kwargs):
        if fam is self._unifamily or fam is None:
            self._subsnap_base._load_array(array_name, self._unifamily)

    def _derive_array(self, array_name, fam=None):
        if fam is self._unifamily or fam is None:
            self._subsnap_base._derive_array(array_name, self._unifamily)



from . import (
    cosmology,
    halo,
    hifrac,
    hmf,
    interpolate,
    ionfrac,
    luminosity,
    pkdgrav_cosmo,
    profile,
    ramses_util,
    theoretical_profiles,
)
from .decomp import decomp
from .hmf import halo_mass_function



import numpy as np

import pynbody


def test_order_bridge():
    f1 = pynbody.new(dm=1000)
    f2 = pynbody.new(dm=3000)
    f1['iord'] = np.arange(5, 2005, 2, dtype=int)
    f2['iord'] = np.arange(3000, dtype=int)

    b = pynbody.bridge.OrderBridge(f1, f2)

    h1 = f1[:50:2]
    assert b(h1).ancestor is f2
    assert (b(h1)['iord'] == h1['iord']).all


def test_bridge_factory():
    f1 = pynbody.load("testdata/g15784.lr.01024")
    f2 = pynbody.load("testdata/g15784.lr.01024")
    b = pynbody.bridge.bridge_factory(f1, f2)

    assert isinstance(b, pynbody.bridge.OrderBridge)


def test_nonmonotonic_transfer_matrix():
    # test the case where a non-monotonic bridge between two simulations is queries for particle transfer
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=10)
    f1['iord'] = np.arange(0,10)
    f2['iord'] = np.array([0,2,4,6,8,1,3,5,7,9],dtype=np.int32)


    # 100% of mass in group 1 transfers to group 1
    # 80% of mass in group 0 transfers to group 0
    # 20% of mass in group 0 transfers to group 1

    f1['grp'] = np.array([0,0,0,0,0,1,1,1,1,1],dtype=np.int32)
    f2['grp'] = np.array([0,0,0,0,1,1,1,1,1,1],dtype=np.int32)[f2['iord']]

    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False)

    h1 = pynbody.halo.number_array.HaloNumberCatalogue(f1)
    h2 = pynbody.halo.number_array.HaloNumberCatalogue(f2)

    xfer = b.catalog_transfer_matrix(0,1,h1,h2)

    assert (xfer==[[4,1],[0,5]]).all()


def test_nonmonotonic_incomplete_bridge():
    # Test the non-monotonic bridge where not all the particles are shared
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=9)
    f1['iord'] = np.arange(0,10)
    f2['iord'] = np.array([0,2,4,6,8,1,5,7,9])
    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False)

    assert (b(f1[:5])['iord']==np.array([0,1,2,4])).all()


def test_monotonic_incomplete_bridge():
    # Test the monotonic bridge where not all the particles are shared
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=9)
    f1['iord'] = np.arange(0, 10)
    f2['iord'] = np.array([0, 1, 2, 4, 5 ,6, 7, 8, 9])
    b = pynbody.bridge.OrderBridge(f1, f2, monotonic=False)

    assert (b(f1[:5])['iord'] == np.array([0, 1, 2, 4])).all()


def test_family_morphing():
    f1 = pynbody.new(dm=5, gas=5)
    f2 = pynbody.new(dm=10)

    # set dm and gas iords separately as it's possible the new command initialises them out of order
    f1.dm['iord'] = np.arange(0,5)
    f1.gas['iord'] = np.arange(5,10)
    f2['iord'] = np.array([0,2,4,6,8,1,3,5,7,9])

    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False,allow_family_change=True)

    assert (b(f2).dm['iord']==np.array([0,2,4,1,3])).all()
    assert (b(f2).gas['iord'] == np.array([6, 8, 5, 7, 9])).all()


def test_bridging_with_more_families():
    # Test that we can create a group array for snapshots that have a complex family structure,
    # and bridge only with one family (DM). This is necessary for e.g. Tangos linking
    f1 = pynbody.load("testdata/ramses_new_format_cosmo_with_ahf_output_00110")
    g1 = pynbody.halo.ahf.AHFCatalogue(f1, halo_numbers='v1')
    g1.load_all()
    # Work only on one family and create a useless bridge which is enough to break the code
    f1 = f1.dm
    b = pynbody.bridge.OrderBridge(f1, f1)

    # The line below would fail with IndexError: index is out of bound when called for a single family
    # The family slicing is done wrto to the base ancestor, but the IDs were not offset
    # by family.start and would run over the array. This is now fixed
    mat = b.catalog_transfer_matrix(max_index=5, groups_1=g1, groups_2=g1, use_family=pynbody.family.dm, only_family=pynbody.family.dm)
    # Now it works and produces a diagonla matrix  since we are mapping the same snapshot
    assert(np.count_nonzero(mat - np.diag(np.diagonal(mat))) == 0)
    assert(mat[0, 0] == 12)

def test_fuzzy_match_only_one_family():
    f = pynbody.new(dm=10, gas=10)
    f['iord'] = np.arange(0,20)
    f['grp'] = np.array([0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1],dtype=np.int32)

    f2 = pynbody.new(dm=10, gas=10)
    f2['iord'] = np.arange(0, 20)
    f2['grp'] = f['grp']
    f2.dm['grp'] = np.array([0,0,0,0,0,1,1,1,2,2],dtype=np.int32)

    h = pynbody.halo.number_array.HaloNumberCatalogue(f)
    h2 = pynbody.halo.number_array.HaloNumberCatalogue(f2)
    b = pynbody.bridge.OrderBridge(f,f2)
    assert b.fuzzy_match_catalog(use_family=pynbody.family.gas,groups_1=h, groups_2=h2)[1]==[(1, 1.0)]
    assert b.fuzzy_match_catalog(use_family=pynbody.family.dm, groups_1=h, groups_2=h2)[1] == [(1, 0.6), (2, 0.4)]
    assert b.fuzzy_match_catalog(groups_1=h, groups_2=h2)[1] == [(1, 0.8), (2, 0.2)]

    # Test that it also works with only_family:
    assert b.fuzzy_match_catalog(only_family=pynbody.family.gas, groups_1=h, groups_2=h2)[1] == [(1, 1.0)]
    assert b.fuzzy_match_catalog(only_family=pynbody.family.dm, groups_1=h, groups_2=h2)[1] == [(1, 0.6), (2, 0.4)]
